# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.colors as mcolors
import datastock as ds

# specific
from . import _generic_check
from. import _class1_plot
# from . import _class1_compute as _compute
from . import _spectralunits


# #############################################################################
# #############################################################################
#                           plot mesh
# #############################################################################


def _plot_mesh_prepare_spectral(
    coll=None,
    key=None,
):

    # --------
    # prepare

    kknots = coll.dobj[coll._which_msp][key]['knots'][0]
    knots = coll.ddata[kknots]['data']

    xx = np.array([knots, knots, np.full(knots.shape, np.nan)]).T.ravel()
    yy = np.array([
        np.zeros(knots.shape),
        np.ones(knots.shape),
        np.ones(knots.shape),
    ]).T.ravel()

    return xx, yy


def plot_mesh_spectral(
    coll=None,
    key=None,
    ind_knot=None,
    ind_cent=None,
    units=None,
    nmax=None,
    color=None,
    dax=None,
    dmargin=None,
    fs=None,
    dleg=None,
    connect=None,
):
    """ Plot the desired spectral mesh

    """

    # --------------
    # check input

    (
     key, ind_knot, ind_cent, _, _, color, dleg,
     ) = _class1_plot._plot_mesh_check(
        coll=coll,
        which_mesh=coll._which_msp,
        key=key,
        ind_knot=ind_knot,
        ind_cent=ind_cent,
        color=color,
        dleg=dleg,
    )

    # --------------
    #  Prepare data

    xx, yy = _plot_mesh_prepare_spectral(
        coll=coll,
        key=key,
    )

    if units not in [None, 'eV']:
        xx, _, _, cat = _spectralunits.convert_spectral(
            data_in=xx,
            units_in='eV',
            units_out=units,
        )
        xlab = cat + r" ($" + units + "$)"

    else:
        xlab = r'energy ($eV$)'

    # --------------
    # plot - prepare

    if dax is None:

        if dmargin is None:
            dmargin = {
                'left': 0.1, 'right': 0.9,
                'bottom': 0.1, 'top': 0.9,
                'hspace': 0.1, 'wspace': 0.1,
            }

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(ncols=1, nrows=1, **dmargin)
        ax0 = fig.add_subplot(gs[0, 0])
        ax0.set_xlabel(xlab)

        dax = {'spectral': ax0}

    dax = _generic_check._check_dax(dax=dax, main='spectral')

    # --------------
    # plot

    kax = 'spectral'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ax.plot(
            xx,
            yy,
            ls='-',
            lw=0.5,
            color=color,
            alpha=0.5,
            label=key,
        )

        if ind_knot is not None:
            ax.plot(
                ind_knot[0],
                0.5,
                marker='o',
                ms=8,
                ls='None',
                color=color,
                label='knots',
            )
            # ax.plot(
            #     ind_knot[1][0, :, :],
            #     marker='x',
            #     ms=4,
            #     ls='None',
            #     color=color,
            # )

        if ind_cent is not None:
            ax.plot(
                ind_cent[0],
                0.5,
                marker='x',
                ms=8,
                ls='None',
                color=color,
                label='cents',
            )
            # ax.plot(
            #     ind_cent[1][0, :, :],
            #     marker='o',
            #     ms=4,
            #     ls='None',
            #     color=color,
            # )

    # --------------
    # dleg

    if dleg is not False:
        for kax in dax.keys():
            dax[kax]['handle'].legend(**dleg)

    return dax


# #############################################################################
# #############################################################################
#                           plot bspline
# #############################################################################


def _plot_bspline_prepare_spectral(
    coll=None,
    key=None,
    keym=None,
    res=None,
    mode=None,
):

    xx = coll.get_sample_mesh_spectral(key=keym, res=res, mode=mode)

    # TBF depending on what is practical for LOS
    yy = coll.interpolate_spectrum(
        key=None,
        coefs=None,
        E=xx,
        t=None,
        indt=None,
        indt_strict=None,
        indbs=None,
        details=True,
        reshape=None,
        res=None,
        nan0=None,
        val_out=None,
        return_params=False,
        store=False,
        inplace=False,
    )

    return xx, yy


def plot_bspline_spectral(
    # ressources
    coll=None,
    # inputs
    key=None,
    indbs=None,
    indt=None,
    # parameters
    knots=None,
    cents=None,
    res=None,
    plot_mesh=None,
    val_out=None,
    nan0=None,
    # plot-specific
    cmap=None,
    dax=None,
    dmargin=None,
    fs=None,
    dleg=None,
):

    # --------------
    # check input

    (
        key, _, keym, mtype0, mtype,
        indbs, indt,
        knots, cents, knotsi, centsi,
        plot_mesh, _, dleg,
    ) = _class1_plot._plot_bspline_check(
        coll=coll,
        key=key,
        indbs=indbs,
        indt=indt,
        knots=knots,
        cents=cents,
        plot_mesh=plot_mesh,
        dleg=dleg,
    )

    # --------------
    #  Prepare data

    xx, yy = _plot_bspline_prepare_spectral(
        coll=coll,
        key=key,
        keym=keym,
        mtype0=mtype0,
        indbs=indbs,
        indt=indt,
        knotsi=knotsi,
        centsi=centsi,
        res=res,
        val_out=val_out,
        nan0=nan0,
    )

    # --------------
    # plot - prepare

    if dax is None:

        if dmargin is None:
            dmargin = {
                'left': 0.1, 'right': 0.9,
                'bottom': 0.1, 'top': 0.9,
                'hspace': 0.1, 'wspace': 0.1,
            }

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(ncols=1, nrows=1, **dmargin)
        ax0 = fig.add_subplot(gs[0, 0], aspect='equal')
        ax0.set_xlabel(f'R (m)')
        ax0.set_ylabel(f'Z (m)')

        dax = {'spectral': ax0}

    dax = _generic_check._check_dax(dax=dax, main='spectral')

    # --------------
    # plot

    kax = 'spectral'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ax.plot(
            xx,
            np.sum(yy, axis=0),
            c='k',
            lw=2.,
            ls='-',
        )

        ax.plot(
            xx,
            yy.T,
            lw=1.,
            ls='-',
        )

        if knots is not False:
            ax.plot(
                knotsi[0].ravel(),
                knotsi[1].ravel(),
                marker='x',
                ms=6,
                ls='None',
                color='k',
            )

        if cents is not False:
            ax.plot(
                centsi[0].ravel(),
                centsi[1].ravel(),
                marker='o',
                ms=6,
                ls='None',
                color='k',
            )

        ax.relim()
        ax.autoscale()

        # --------------
        # dleg

        if dleg is not False:
            ax.legend(**dleg)

    return dax


# #############################################################################
# #############################################################################
#                           plot spectrum
#               (1d or time-dependent and/or radius dependent)
# #############################################################################


def plot_spectrum(
    coll=None,
    key=None,
):

    # -----------
    # check input

    key, keyX, keyY = None, None, None

    # ------------------------
    # call appropriate routine

    return coll.plot_as_array(
        key=key,
        keyX=keyX,
        keyY=keyY,
    )


# #############################################################################
# #############################################################################
#                           plot profile2d
# #############################################################################


def _plot_profile2d_check(
    coll=None,
    key=None,
    cmap=None,
    dcolorbar=None,
    dleg=None,
):

    # key
    dk = coll.get_profiles2d()
    key = ds._generic_check._check_var(
        key, 'key', types=str, allowed=list(dk.keys())
    )
    keybs = dk[key]
    refbs = coll.dobj['bsplines'][keybs]['ref']
    keym = coll.dobj['bsplines'][keybs]['mesh']
    mtype = coll.dobj[coll._which_mesh][keym]['type']

    # cmap
    if cmap is None:
        cmap = 'viridis'

    # dcolorbar
    defdcolorbar = {
        # 'location': 'right',
        'fraction': 0.15,
        'orientation': 'vertical',
    }
    dcolorbar = ds._generic_check._check_var(
        dcolorbar, 'dcolorbar',
        default=defdcolorbar,
        types=dict,
    )

    # dleg
    defdleg = {
        'bbox_to_anchor': (1.1, 1.),
        'loc': 'upper left',
        'frameon': True,
    }
    dleg = ds._generic_check._check_var(
        dleg, 'dleg',
        default=defdleg,
        types=(bool, dict),
    )

    # polar1d
    polar1d = False
    if mtype  == 'polar':
        if coll.dobj['bsplines'][keybs]['class'].knotsa is None:
            polar1d = True
        elif len(coll.dobj['bsplines'][keybs]['ref']) == 2:
            polar1d = True
        else:
            apbs = coll.dobj['bsplines'][keybs]['class'].apex_per_bs_a
            if np.sum([aa is not None for aa in apbs]) == 1:
                polar1d = True

    return key, keybs, keym, cmap, dcolorbar, dleg, polar1d


def _plot_profiles2d_prepare(
    coll=None,
    key=None,
    keybs=None,
    keym=None,
    coefs=None,
    indt=None,
    res=None,
    mtype=None,
):

    # check input
    deg = coll.dobj['bsplines'][keybs]['deg']

    # get dR, dZ
    # dR, dZ, Rminmax, Zminmax = _plot_bsplines_get_dRdZ(
        # coll=coll, km=keym, meshtype=mtype,
    # )
    dR, dZ, Rminmax, Zminmax = None, None, None, None

    if res is None:
        res_coef = 0.2
        res = [res_coef*dR, res_coef*dZ]

    # compute
    coll2 = coll.interpolate_profile2d(
        key=key,
        R=None,
        Z=None,
        coefs=coefs,
        indt=indt,
        res=res,
        details=False,
        nan0=True,
        imshow=False,
        return_params=False,
        store=True,
        inplace=False,
    )

    keymap = [k0 for k0, v0 in coll2.ddata.items() if v0['data'].ndim > 1][0]
    ndim = coll2.ddata[keymap]['data'].ndim

    refmap = coll2.ddata[keymap]['ref']
    dkeys = {
        'key': keymap,
        'keyX': coll2.get_ref_vector(key=keymap, ref=refmap[-2])[3],
        'keyY': coll2.get_ref_vector(key=keymap, ref=refmap[-1])[3],
        'keyZ': None,
    }
    if ndim == 3:
        keyZ = coll2.get_ref_vector(key=keymap, ref=refmap[0])[3]
        import datastock as ds
        uniform = ds._plot_as_array._check_uniform_lin(
            k0=keyZ, ddata=coll2.ddata,
        )
        if not uniform:
            keyZ = None

    if deg == 0:
        interp = 'nearest'
    elif deg == 1:
        interp = 'bilinear'
    elif deg >= 2:
        interp = 'bicubic'

    # radial of polar
    if mtype == 'polar':
        # lcol
        lcol = ['k', 'r', 'b', 'g', 'm', 'c', 'y']

    else:
        lcol = None

    return coll2, dkeys, interp, lcol


def _plot_profile2d_polar_add_radial(
    coll=None,
    key=None,
    keym=None,
    keybs=None,
    dax=None,
):

    # key to radius
    kr2d = coll.dobj[coll._which_mesh][keym]['radius2d']
    kr = coll.dobj[coll._which_mesh][keym]['knots'][0]
    rr = coll.ddata[kr]['data']
    rad = np.linspace(rr[0], rr[-1], rr.size*20)

    # get angle if any
    clas = coll.dobj['bsplines'][keybs]['class']
    if clas.knotsa is None:
        angle = None
    elif len(clas.shapebs) == 2:
        ka = coll.dobj['bsplines'][keybs]['apex'][1]
        angle = coll.ddata[ka]['data']
    elif np.sum(clas.nbs_a_per_r > 1) == 1:
        i0 = (clas.nbs_a_per_r > 1).nonzero()[0][0]
        angle = coll.dobj['bsplines'][keybs]['class'].apex_per_bs_a[i0]
    else:
        pass

    if angle is None:
        radmap = rad
        anglemap = angle
    else:
        radmap = np.repeat(rad[:, None], angle.size, axis=1)
        anglemap = np.repeat(angle[None, :], rad.size, axis=0)

    # reft
    reft, keyt, _, dind = coll.get_time_common(keys=[key, kr2d])[1:]

    # radial total profile
    radial, t_radial, _ = coll.interpolate_profile2d(
        key=key,
        radius=radmap,
        angle=anglemap,
        grid=False,
        t=keyt,
    )

    if reft is not None and radial.ndim == radmap.ndim:
        radial = np.repeat(radial[None, ...], t_radial.size, axis=0)

    # details for purely-radial cases
    if clas.knotsa is None:
        radial_details, t_radial, _ = coll.interpolate_profile2d(
            key=keybs,
            radius=rad,
            angle=None,
            grid=False,
            details=True,
        )

        if reft is None:
            radial_details = radial_details * coll.ddata[key]['data'][None, :]
            refdet = ('nradius',)
        else:
            refdet = (reft, 'nradius')
            if reft == coll.get_time(key)[2]:
                radial_details = (
                    radial_details[None, :, :]
                    * coll.ddata[key]['data'][:, None, :]
                )
            elif key in dind.keys():
                radial_details = (
                    radial_details[None, :, :]
                    * coll.ddata[key]['data'][dind[key]['ind'], None, :]
                )

        nbs = radial_details.shape[-1]

    # add to dax
    dax.add_ref(key='nradius', size=rad.size)
    if angle is not None:
        dax.add_ref(key='nangle', size=angle.size)

    if reft is not None:
        assert radial.ndim > 1 and radial.shape[0] > 1
        # if angle is not None:
            # ref = (reft, 'nangle', 'nradius')
        # else:
        ref = (reft, 'nradius')
    else:
        # if angle is not None:
            # ref = ('nangle', 'nradius')
        # else:
        ref = 'nradius'

    # add to ddata
    kradius = 'radius'
    dax.add_data(key=kradius, data=rad, ref='nradius')
    if angle is None:
        lk = ['radial']
        dax.add_data(key=lk[0], data=radial, ref=ref)
        lkdet = [f'radial-detail-{ii}' for ii in range(nbs)]
        for ii in range(nbs):
            dax.add_data(
                key=lkdet[ii], data=radial_details[..., ii], ref=refdet,
            )

    else:
        kangle = 'angle'
        dax.add_data(key=kangle, data=angle, ref='nangle')
        lkdet = None
        lk = [f'radial-{ii}' for ii in range(angle.size)]
        if reft is None:
            for ii in range(angle.size):
                dax.add_data(key=lk[ii], data=radial[:, ii], ref=ref)
        else:
            for ii in range(angle.size):
                dax.add_data(key=lk[ii], data=radial[:, :, ii], ref=ref)

    return kradius, lk, lkdet, reft


def plot_profile2d(
    # ressources
    coll=None,
    # inputs
    key=None,
    # parameters
    coefs=None,
    indt=None,
    res=None,
    # figure
    vmin=None,
    vmax=None,
    cmap=None,
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
    # interactivity
    dinc=None,
    connect=None,
):

    # --------------
    # check input

    if connect is None:
        connect = True

    (
        key, keybs, keym, cmap, dcolorbar, dleg, polar1d,
    ) = _plot_profile2d_check(
        coll=coll,
        key=key,
        cmap=cmap,
        dcolorbar=dcolorbar,
        dleg=dleg,
    )
    mtype = coll.dobj[coll._which_mesh][keym]['type']
    hastime = coll.get_time(key=key)[0]

    # --------------
    #  Prepare data

    (
        coll2, dkeys, interp, lcol,
    ) = _plot_profiles2d_prepare(
        coll=coll,
        key=key,
        keybs=keybs,
        keym=keym,
        coefs=coefs,
        indt=indt,
        res=res,
        mtype=mtype,
    )

    # ---------------
    # call right function

    if mtype in ['rect', 'tri']:
        return coll2.plot_as_array(
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
            interp=interp,
            connect=connect,
            **dkeys,
        )

    else:

        if dax is None:
            dax = _plot_profile2d_polar_create_axes(
                fs=fs,
                dmargin=dmargin,
            )

        dax, dgroup = coll2.plot_as_array(
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
            connect=False,
            interp=interp,
            label=True,
            **dkeys,
        )

        # ------------------
        # add radial profile to dax

        kradius, lkradial, lkdet, reft = _plot_profile2d_polar_add_radial(
            coll=coll,
            key=key,
            keym=keym,
            keybs=keybs,
            dax=dax,
        )


        assert (reft is not None) == ('Z' in dgroup.keys())
        if reft is not None and reft not in dgroup['Z']['ref']:
            dgroup['Z']['ref'].append(reft)
            dgroup['Z']['data'].append('index')

        # ------------------
        # add radial profile

        kax = 'radial'
        if dax.dax.get(kax) is not None:
            ax = dax.dax[kax]['handle']
            for ii in range(len(lkradial)):

                if reft is None:
                    l0, = ax.plot(
                        dax.ddata[kradius]['data'],
                        dax.ddata[lkradial[ii]]['data'],
                        c=lcol[ii],
                        ls='-',
                        lw=2,
                    )
                else:
                    l0, = ax.plot(
                        dax.ddata[kradius]['data'],
                        dax.ddata[lkradial[ii]]['data'][0, :],
                        c=lcol[ii],
                        ls='-',
                        lw=2,
                    )

                    kl = f"radial{ii}"
                    dax.add_mobile(
                        key=kl,
                        handle=l0,
                        refs=(reft,),
                        data=[lkradial[ii]],
                        dtype=['ydata'],
                        axes=kax,
                        ind=0,
                    )

            if lkdet is not None:
                for ii in range(len(lkdet)):
                    if reft is None:
                        l0, = ax.plot(
                            dax.ddata[kradius]['data'],
                            dax.ddata[lkdet[ii]]['data'],
                            ls='-',
                            lw=1,
                        )
                    else:
                        l0, = ax.plot(
                            dax.ddata[kradius]['data'],
                            dax.ddata[lkdet[ii]]['data'][0, :],
                            ls='-',
                            lw=1,
                        )

                        kl = f"radial_det{ii}"
                        dax.add_mobile(
                            key=kl,
                            handle=l0,
                            refs=(reft,),
                            data=[lkdet[ii]],
                            dtype=['ydata'],
                            axes=kax,
                            ind=0,
                        )

            ax.set_xlim(
                dax.ddata[kradius]['data'].min(),
                dax.ddata[kradius]['data'].max(),
            )

            if vmin is not None:
                ax.set_ylim(bottom=vmin)
            if vmax is not None:
                ax.set_ylim(top=vmax)

        # connect
        if connect is True:
            dax.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)
            dax.disconnect_old()
            dax.connect()

            dax.show_commands()
            return dax
        else:
            return dax, dgroup



def _plot_profile2d_polar_create_axes(
    fs=None,
    dmargin=None,
):

    if fs is None:
        fs = (15, 9)

    if dmargin is None:
        dmargin = {
            'left': 0.05, 'right': 0.95,
            'bottom': 0.05, 'top': 0.95,
            'hspace': 0.4, 'wspace': 0.3,
        }

    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(ncols=6, nrows=6, **dmargin)

    # axes for image
    ax0 = fig.add_subplot(gs[:4, 2:4], aspect='auto')

    # axes for vertical profile
    ax1 = fig.add_subplot(gs[:4, 4], sharey=ax0)

    # axes for horizontal profile
    ax2 = fig.add_subplot(gs[4:, 2:4], sharex=ax0)

    # axes for traces
    ax3 = fig.add_subplot(gs[2:4, :2])

    # axes for traces
    ax7 = fig.add_subplot(gs[:2, :2], sharey=ax2)

    # axes for text
    ax4 = fig.add_subplot(gs[:3, 5], frameon=False)
    ax5 = fig.add_subplot(gs[3:, 5], frameon=False)
    ax6 = fig.add_subplot(gs[4:, :2], frameon=False)

    # dax
    dax = {
        # data
        'matrix': {'handle': ax0, 'type': 'matrix'},
        'vertical': {'handle': ax1, 'type': 'misc'},
        'horizontal': {'handle': ax2, 'type': 'misc'},
        'traces': {'handle': ax3, 'type': 'misc'},
        'radial': {'handle': ax7, 'type': 'misc'},
        # text
        'textX': {'handle': ax4, 'type': 'text'},
        'textY': {'handle': ax5, 'type': 'text'},
        'textZ': {'handle': ax6, 'type': 'text'},
    }
    return dax



# -*- coding: utf-8 -*-


# Built-in
import copy


# Common
import numpy as np


# local
from ._class02_BSplines2D import BSplines2D as Previous
from . import _class03_checks as _checks
from . import _class03_binning as _binning


__all__ = ['Bins']


# #############################################################################
# #############################################################################
#
# #############################################################################


class Bins(Previous):

    _which_bins = 'bins'
    _ddef = copy.deepcopy(Previous._ddef)
    _dshow = dict(Previous._dshow)

    _dshow.update({
        _which_bins: [
            'nd',
            'cents',
            'shape',
            'ref',
        ],
    })

    # -----------------
    # bsplines
    # ------------------

    def add_bins(
        self,
        key=None,
        edges=None,
        # custom names
        key_ref=None,
        key_cents=None,
        key_res=None,
        # attributes
        **kwdargs,
    ):
        """ Add bin """

        # --------------
        # check inputs

        key, dref, ddata, dobj = _checks.check(
            coll=self,
            key=key,
            edges=edges,
            # custom names
            key_cents=key_cents,
            key_ref=key_ref,
            # attributes
            **kwdargs,
        )

        # --------------
        # update dict and crop if relevant

        self.update(dobj=dobj, ddata=ddata, dref=dref)

    def remove_bins(
        self,
        key=None,
        propagate=None,
    ):

        _checks.remove_bins(
            coll=self,
            key=key,
            propagate=propagate,
        )

    # -----------------
    # binning tools
    # ------------------

    def binning(
        self,
        data=None,
        data_units=None,
        axis=None,
        # binning
        bins0=None,
        bins1=None,
        bin_data0=None,
        bin_data1=None,
        bin_units0=None,
        # kind of binning
        integrate=None,
        statistic=None,
        # options
        safety_ratio=None,
        dref_vector=None,
        verb=None,
        returnas=None,
        # storing
        store=None,
        store_keys=None,
    ):
        """ Bin data along ref_key

        Binning is treated here as an integral
        Hence, if:
            - the data has units [ph/eV]
            - the ref_key has units [eV]
            - the binned data has units [ph]

        return a dict with data and units per key

        """

        return _binning.binning(
            coll=self,
            data=data,
            data_units=data_units,
            axis=axis,
            # binning
            bins0=bins0,
            bins1=bins1,
            bin_data0=bin_data0,
            bin_data1=bin_data1,
            bin_units0=bin_units0,
            # kind of binning
            integrate=integrate,
            statistic=statistic,
            # options
            safety_ratio=safety_ratio,
            dref_vector=dref_vector,
            verb=verb,
            returnas=returnas,
            # storing
            store=store,
            store_keys=store_keys,
        )


# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import scipy.interpolate as scpinterp
from matplotlib.tri import Triangulation as mplTri
import datastock as ds


# specific
from . import _class01_checks_2d_tri as _checks
# from . import _class02_bsplines_operators_tri
from . import _class02_bsplines_rect as _mbr


# #############################################################################
# #############################################################################
#                       BivariateSplineRect - scipy subclass
# #############################################################################


class BivariateSplineTri(scpinterp.BivariateSpline):
    """ Subclass for tofu

    Defined from knots (unique) and deg
    coefs set to 1 by default

    """

    def __init__(
        self,
        knots0=None,
        knots1=None,
        indices=None,
        deg=None,
    ):
        """ Class handling triangular bsplines """

        # ------------
        # check inputs

        knots = np.array([knots0, knots1]).T
        indices, knots = _checks._mesh2DTri_conformity(
            knots=knots, indices=indices, key='class',
        )
        indices = _checks._mesh2DTri_ccw(
            knots=knots, indices=indices, key='class',
        )

        # get trifinder
        trifind = mplTri(knots0, knots1, indices).get_trifinder()

        # deg
        deg = ds._generic_check._check_var(
            deg, 'deg',
            types=int,
            default=1,
            allowed=[0, 1],
        )

        self.knots0 = knots0
        self.knots1 = knots1
        self.nknots = knots0.size
        self.indices = indices
        self.nind = indices.shape[0]
        self.deg = deg
        self.trifind = trifind

        # ------------
        # get neigh cents per knot

        self.cents_per_knots = self._get_cents_per_knots()

        # ------------
        # get height per cent /  knot

        self.heights = self._get_heights_per_centsknots()

        # ------------
        # nbsplines

        if deg == 0:
            nbs = self.nind
        elif deg == 1:
            nbs = self.nknots
        elif deg == 2:
            raise NotImplementedError()

        self.nbs = nbs
        self.shapebs = (nbs,)
        self.coefs = np.ones((nbs,), dtype=float)

    def _get_cents_per_knots(self):
        """ Return a (nknots, nmax) array of int indices

        Array contains -1 where there is no cent anymore
        """

        out = [
            np.any(self.indices == ii, axis=1).nonzero()[0]
            for ii in range(self.nknots)
        ]
        nmax = np.array([oo.size for oo in out])

        cents_per_knots = -np.ones((self.nknots, nmax.max()), dtype=int)
        for ii in range(self.nknots):
            cents_per_knots[ii, :nmax[ii]] = out[ii]
        return cents_per_knots

    def _get_heights_per_centsknots(self):
        """ Return the height of each knot in each cent

        Returnad as (ncents, 3) array, like cents
        """

        x0 = self.knots0[self.indices]
        x1 = self.knots1[self.indices]

        heights = np.full(self.indices.shape, np.nan)

        for iref, (i0, i1) in enumerate([(1, 2), (2, 0), (0, 1)]):
            base = np.sqrt(
                (x0[:, i1] - x0[:, i0])**2 + (x1[:, i1] - x1[:, i0])**2
            )
            heights[:, iref] = np.abs(
                (x0[:, i0] - x0[:, iref])*(x1[:, i1] - x1[:, iref])
                - (x1[:, i0] - x1[:, iref])*(x0[:, i1] - x0[:, iref])
            ) / base

        return heights

    def get_heights_per_centsknots_pts(self, x, y):
        """ Return the height of each knot in each cent

        Returnad as (ncents, 3) array, like cents

        OPTIMIZATION POSSIBLE FOR EV_DETAILS BY TAKING INDBS AS INPUT ARG !!!
        """

        if x.shape != y.shape:
            msg = "Arg x and y must have the same shape!"
            raise Exception(msg)

        x0 = self.knots0[self.indices]
        x1 = self.knots1[self.indices]

        heights = np.full(tuple(np.r_[x.shape, 3]), np.nan)
        ind = self.trifind(x, y)

        for ii in np.unique(ind):

            if ii == -1:
                continue
            indi = ind == ii

            for iref, (i0, i1) in enumerate([(1, 2), (2, 0), (0, 1)]):
                v_base = np.array([
                    x0[ii, i1] - x0[ii, i0],
                    x1[ii, i1] - x1[ii, i0],
                ])
                v_perp = np.array([v_base[1], -v_base[0]])
                v_base = v_base / np.linalg.norm(v_base)
                v_perp = v_perp / np.linalg.norm(v_perp)

                v0 = np.array([
                    x0[ii, i0] - x0[ii, iref],
                    x1[ii, i0] - x1[ii, iref],
                ])
                v0_base = v0[0]*v_base[0] + v0[1]*v_base[1]
                v0_perp = v0[0]*v_perp[0] + v0[1]*v_perp[1]

                v_height = (v0 + (-v0_base*v_base + v0_perp*v_perp))/2.
                v_height_norm = np.linalg.norm(v_height)

                dx0 = x[indi] - x0[ii, iref]
                dx1 = y[indi] - x1[ii, iref]
                heights[indi, iref] = (
                    dx0*v_height[0] + dx1*v_height[1]
                ) / v_height_norm**2

        indok = ~np.isnan(heights)
        assert np.all(heights[indok] >= 0. - 1e-14)
        assert np.all(heights[indok] <= 1. + 1e-14)
        return heights, ind

    # --------
    # bsplines

    def _get_knotscents_per_bs(
        self,
        ind=None,
        return_cents=None,
        return_knots=None,
        returnas=None,
    ):
        """ Return 2 arrays of int indices

        A (nbs, ncents_per_bs) array
        A (nbs, nknots_per_bs) array
        """

        # ------------
        # check inputs

        return_cents = ds._generic_check._check_var(
            return_cents, 'return_cents',
            types=bool,
            default=True,
        )

        return_knots = ds._generic_check._check_var(
            return_knots, 'return_knots',
            types=bool,
            default=True,
        )

        returnas = ds._generic_check._check_var(
            returnas, 'returnas',
            types=str,
            allowed=['ind', 'data'],
            default='ind',
        )

        if ind is None:
            ind = np.ones((self.nbs,), dtype=bool)
        ind_num = ind.nonzero()[0]
        nbs = ind.sum()

        # ------------
        # added for details

        if self.deg == 0:
            if return_cents:
                cents_per_bs = ind_num[:, None]
            if return_knots:
                knots_per_bs = self.indices[ind, :]

        elif self.deg == 1:
            if return_cents or return_knots:
                cents_per_bs = self.cents_per_knots[ind, :]
            if return_knots:
                nmax_cents = np.sum(cents_per_bs >= 0, axis=1)
                nmax = self.cents_per_knots.shape[1] + 3
                knots_per_bs = -np.ones((nbs, nmax), dtype=int)
                knots_per_bs[:, 0] = ind_num
                for ii, i0 in enumerate(ind_num):
                    nu = np.unique(
                        self.indices[cents_per_bs[ii, :nmax_cents[ii]], :]
                    )
                    knots_per_bs[ii, 1:nu.size] = [nn for nn in nu if nn != i0]

        elif self.deg == 2:
            raise NotImplementedError()

        # ------
        # return

        if returnas == 'data':

            # cents
            if return_cents:
                nmax = np.sum(cents_per_bs >= 0, axis=1)
                cents_per_bs_temp = np.full((2, nbs, nmax.max()), np.nan)
                for ii in range(nbs):
                    ind_temp = self.indices[cents_per_bs[ii, :nmax[ii]], :]
                    cents_per_bs_temp[0, ii, :nmax[ii]] = np.mean(
                        self.knots0[ind_temp],
                        axis=1,
                    )
                    cents_per_bs_temp[1, ii, :nmax[ii]] = np.mean(
                        self.knots1[ind_temp],
                        axis=1,
                    )
                cents_per_bs = cents_per_bs_temp

            # knots
            if return_knots:
                nmax = np.sum(knots_per_bs >= 0, axis=1)
                knots_per_bs_temp = np.full((2, nbs, nmax.max()), np.nan)
                for ii in range(nbs):
                    ind_temp = knots_per_bs[ii, :nmax[ii]]
                    knots_per_bs_temp[0, ii, :nmax[ii]] = self.knotsR[ind_temp]
                    knots_per_bs_temp[1, ii, :nmax[ii]] = self.knotsZ[ind_temp]
                knots_per_bs = knots_per_bs_temp

        # return
        if return_cents and return_knots:
            return knots_per_bs, cents_per_bs
        elif return_cents:
            return cents_per_bs
        elif return_knots:
            return knots_per_bs

    def _get_bs_cents(
        self,
        ind=None,
    ):
        """ Return (2, nbs) array of cordinates of centers per bspline

        """

        # ------------
        # check inputs

        if ind is None:
            ind = np.ones((self.nbs,), dtype=bool)

        # ------------
        # added for details

        if self.deg == 0:
            bs_cents = np.array([
                np.mean(self.knots0[self.indices[ind, :]], axis=1),
                np.mean(self.knots1[self.indices[ind, :]], axis=1),
            ])

        elif self.deg == 1:
            bs_cents = np.array([
                self.knots0[ind],
                self.knots1[ind],
            ])

        elif self.deg == 2:
            raise NotImplementedError()

        return bs_cents

    # --------
    # evaluation checks

    def _check_coefs(self, coefs=None):
        """ None for ev_details, (nt, shapebs) for sum """
        if coefs is not None:
            assert coefs.ndim == len(self.shapebs) + 1
            assert coefs.shape[1:] == self.shapebs

    def _ev_generic(
        self,
        indbs=None,
    ):
        # -----------
        # check input

        if indbs is None:
            indbs = np.ones((self.nbs,), dtype=bool)
        else:
            indbs = np.atleast_1d(indbs).ravel()

        c0 = (
            isinstance(indbs, np.ndarray)
            and (
                ('bool' in indbs.dtype.name and indbs.size == self.nbs)
                or ('int' in indbs.dtype.name)
            )
        )
        if not c0:
            msg = (
                "Arg indbs must be  a (nbs,) bool or int array!"
                "\nProvided: {indbs}"
            )
            raise Exception(msg)

        if 'bool' in indbs.dtype.name:
            indbs = indbs.nonzero()[0]

        # indbs => indcent : triangles which are ok
        knots_per_bs, cents_per_bs = self._get_knotscents_per_bs(
            return_cents=True,
            return_knots=True,
            returnas='ind',
        )
        indcent = np.unique(cents_per_bs[indbs, :])
        indcent = indcent[indcent >= 0]

        # -----------
        # prepare

        nbs = indbs.size

        return nbs, knots_per_bs, cents_per_bs, indcent

    # --------
    # evaluation

    def ev_details(
        self,
        x0=None,
        x1=None,
        indbs_tf=None,
        # for compatibility (unused)
        **kwdargs,
    ):

        # -----------
        # generic

        # parameters
        (
            nbs, knots_per_bs, cents_per_bs, indcent,
        ) = self._ev_generic(indbs=indbs_tf)

        # -----------
        # compute

        val = np.zeros(tuple(np.r_[x0.shape, nbs]))
        heights, ind = self.get_heights_per_centsknots_pts(x0, x1)

        indu = np.unique(ind[ind >= 0])
        if self.deg == 0:
            if indbs_tf is None:
                for ii in np.intersect1d(indu, indcent):
                    indi = ind == ii
                    val[indi, ii] = 1.
            else:
                for ii in np.intersect1d(indu, indcent):
                    indi = ind == ii
                    ibs = indbs_tf == ii
                    val[indi, ibs] = 1.

        elif self.deg == 1:
            if indbs_tf is None:
                for ii in np.intersect1d(indu, indcent):
                    indi = ind == ii
                    # get bs
                    ibs = np.any(cents_per_bs == ii, axis=1).nonzero()[0]
                    sorter = np.argsort(self.indices[ii, :])
                    inum = sorter[np.searchsorted(
                        self.indices[ii, :],
                        knots_per_bs[ibs, 0],
                        sorter=sorter,
                    )]
                    for jj, jbs in enumerate(ibs):
                        val[indi, jbs] = 1. - heights[indi, inum[jj]]
            else:
                for ii in np.intersect1d(indu, indcent):
                    indi = ind == ii
                    # get bs
                    ibs = np.intersect1d(
                        indbs_tf,
                        np.any(cents_per_bs == ii, axis=1).nonzero()[0],
                    )
                    sorter = np.argsort(self.indices[ii, :])
                    inum = sorter[np.searchsorted(
                        self.indices[ii, :],
                        knots_per_bs[ibs, 0],
                        sorter=sorter,
                    )]
                    for jj, jbs in enumerate(ibs):
                        ij = indbs_tf == jbs
                        val[indi, ij] = 1. - heights[indi, inum[jj]]
        return val

    def __call__(
        self,
        # interp points
        x0=None,
        x1=None,
        # coefs
        coefs=None,
        axis=None,
        # options
        val_out=None,
        # slicing
        sli_c=None,
        # sli_x=None,
        sli_v=None,
        sli_o=None,
        # indokx0=None,
        shape_v=None,
        shape_o=None,
        axis_v=None,
        dref_com=None,
        # for compatibility (unused)
        **kwdargs,
    ):

        # -----------
        # generic

        # parameters
        (
            nbs, knots_per_bs, cents_per_bs, indcent,
        ) = self._ev_generic(indbs=None)

        # -----------
        # prepare

        val = np.zeros(shape_v)
        heights, ind = self.get_heights_per_centsknots_pts(x0, x1)

        # -----------
        # compute

        indu = np.unique(ind[ind >= 0])
        if self.deg == 0:

            for ii in np.intersect1d(indu, indcent):
                sli_v[axis_v[0]] = (ind == ii)
                sli_c[axis[0]] = [ii]
                val[tuple(sli_v)] += coefs[tuple(sli_c)]

        elif self.deg == 1:
            for ii in np.intersect1d(indu, indcent):

                indi = ind == ii
                sli_v[axis_v[0]] = indi
                shape_height = tuple([
                    indi.sum() if ii in axis else 1
                    for ii in range(len(coefs.shape))
                ])

                # get bs
                ibs = np.any(cents_per_bs == ii, axis=1).nonzero()[0]
                sorter = np.argsort(self.indices[ii, :])
                inum = sorter[np.searchsorted(
                    self.indices[ii, :],
                    knots_per_bs[ibs, 0],
                    sorter=sorter,
                )]

                for jj, jbs in enumerate(ibs):
                    sli_c[axis[0]] = [jbs]

                    if np.prod(shape_height) != heights[indi, inum[jj]].size:
                        import pdb; pdb.set_trace()     # DB
                    val[tuple(sli_v)] += np.reshape(
                        1. - heights[indi, inum[jj]],
                        shape_height,
                    ) * coefs[tuple(sli_c)]

        # clean out-of-mesh
        if dref_com['ix'] is None and val_out is not False:
            slio = sli_o(ind == -1)
            val[slio] = val_out
        return val

    # TBC
    def get_overlap(self):
        raise NotImplementedError()
        return _get_overlap(
            deg=self.degrees[0],
            knotsx=self.knots_per_bs_x,
            knotsy=self.knots_per_bs_y,
            shapebs=self.shapebs,
        )

    # TBD / TBF
    # def get_operator(
        # self,
        # operator=None,
        # geometry=None,
        # cropbs_flat=None,
        # # specific to deg = 0
        # cropbs=None,
    # ):
        # """ Get desired operator """
        # raise NotImplementedError()
        # return _class1_bsplines_operators_tri.get_mesh2dRect_operators(
            # deg=self.degrees[0],
            # operator=operator,
            # geometry=geometry,
            # knotsx_mult=self.tck[0],
            # knotsy_mult=self.tck[1],
            # knotsx_per_bs=self.knots_per_bs_x,
            # knotsy_per_bs=self.knots_per_bs_y,
            # overlap=self.get_overlap(),
            # cropbs_flat=cropbs_flat,
            # cropbs=cropbs,
        # )


# #############################################################################
# #############################################################################
#                       Mesh2Dtri - bsplines - overlap
# #############################################################################


def _get_overlap(
    deg=None,
    knotsx=None,
    knotsy=None,
    shapebs=None,
):
    raise NotImplementedError()


# #############################################################################
# #############################################################################
#                           Mesh2DRect - bsplines
# #############################################################################


def get_bs_class(
    knots0=None,
    knots1=None,
    indices=None,
    deg=None,
):

    # -----------------
    # Instanciate class

    return BivariateSplineTri(
        knots0=knots0,
        knots1=knots1,
        indices=indices,
        deg=deg,
    )



# -*- coding: utf-8 -*-


# Built-in
import copy
import itertools as itt


# Common
import numpy as np
import datastock as ds


# tofu
from . import _utils_bsplines
# from . import _class02_checks as _checks
from . import _class02_bsplines_rect
from . import _class02_bsplines_tri
from . import _class02_bsplines_polar
from . import _class02_bsplines_1d



# ##################################################################
# ##################################################################
#                           Mesh1D - bsplines
# ##################################################################


def _mesh1d_bsplines(
    coll=None,
    keym=None,
    keybs=None,
    deg=None,
):

    kknots = coll.dobj[coll._which_mesh][keym]['knots'][0]
    knots = coll.ddata[kknots]['data']

    kbsn = f'{keybs}_nbs'
    kbsap = f'{keybs}_ap'

    clas = get_bs_class(
        nd='1d',
        deg=deg,
        knots0=knots,
    )

    # ------------
    # refs

    ref = (kbsn,)
    apex = (kbsap,)

    # ----------------
    # format into dict

    # dref
    dref = {
        # bs index
        kbsn: {'size': clas.nbs},
    }

    # ddata
    ddata = {
        kbsap: {
            'data': clas.apex_per_bs,
            'units': coll.ddata[kknots]['units'],
            'dim': coll.ddata[kknots]['dim'],
            'quant': coll.ddata[kknots]['quant'],
            'name': coll.ddata[kknots]['name'],
            'ref': (kbsn,),
        },
    }

    # dobj
    wm = coll._which_mesh
    wbs = coll._which_bsplines
    dobj = {
        wbs: {
            keybs: {
                'deg': deg,
                wm: keym,
                'ref': ref,
                'ref-bs': (kbsn,),
                'apex': apex,
                'shape': clas.shapebs,
                'class': clas,
                'crop': None,
            }
        },
    }

    return dref, ddata, dobj


# ################################################################
# ################################################################
#                           Mesh2DRect - bsplines
# ################################################################


def _mesh2DRect_bsplines(coll=None, keym=None, keybs=None, deg=None):

    # --------------
    # create bsplines

    k0, k1 = coll.dobj[coll._which_mesh][keym]['knots']
    knots0 = coll.ddata[k0]['data']
    knots1 = coll.ddata[k1]['data']

    keybsr = f'{keybs}_nbs'
    kRbsapn = f'{keybs}_n0'
    kZbsapn = f'{keybs}_n1'
    kRbsap = f'{keybs}_ap0'
    kZbsap = f'{keybs}_ap1'

    (
        shapebs, Rbs_apex, Zbs_apex,
        knots_per_bs_R, knots_per_bs_Z,
    ) = _class02_bsplines_rect.get_bs2d_x01(
        deg=deg, knots0=knots0, knots1=knots1,
    )
    nbs = int(np.prod(shapebs))

    clas = get_bs_class(
        nd='2d',
        mtype='rect',
        deg=deg,
        knots0=knots0,
        knots1=knots1,
        shapebs=shapebs,
    )

    # ----------------
    # format into dict

    dref = {
        kRbsapn: {
            'size': Rbs_apex.size,
        },
        kZbsapn: {
            'size': Zbs_apex.size,
        },
        keybsr: {
            'size': nbs,
        },
    }

    ddata = {
        kRbsap: {
            'data': Rbs_apex,
            'units': coll.ddata[k0]['units'],
            'dim': coll.ddata[k0]['dim'],
            'quant': coll.ddata[k0]['quant'],
            'name': coll.ddata[k0]['name'],
            'ref': kRbsapn,
        },
        kZbsap: {
            'data': Zbs_apex,
            'units': coll.ddata[k1]['units'],
            'dim': coll.ddata[k1]['dim'],
            'quant': coll.ddata[k1]['quant'],
            'name': coll.ddata[k1]['name'],
            'ref': kZbsapn,
        },
    }

    dobj = {
        'bsplines': {
            keybs: {
                'deg': deg,
                'mesh': keym,
                'ref': (kRbsapn, kZbsapn),
                'ref-bs': (keybsr,),
                'apex': (kRbsap, kZbsap),
                'shape': shapebs,
                'crop': False,
                'class': clas,
            }
        },
    }

    return dref, ddata, dobj


# ##################################################################
# ##################################################################
#                           Mesh2 - Tri - bsplines
# ##################################################################


def _mesh2DTri_bsplines(coll=None, keym=None, keybs=None, deg=None):

    # --------------
    # create bsplines

    wm = coll._which_mesh
    kknots = coll.dobj[wm][keym]['knots']
    clas = get_bs_class(
        nd='2d',
        mtype='tri',
        deg=deg,
        knots0=coll.ddata[kknots[0]]['data'],
        knots1=coll.ddata[kknots[1]]['data'],
        indices=coll.ddata[coll.dobj[wm][keym]['ind']]['data'],
    )
    keybsr = f'{keybs}_nbs'
    kbscr = f'{keybs}_ap0'
    kbscz = f'{keybs}_ap1'

    bs_cents = clas._get_bs_cents()

    # ----------------
    # format into dict

    dref = {
        # bs index
        keybsr: {
            'size': clas.nbs,
        },
    }

    ddata = {
        kbscr: {
            'data': bs_cents[0, :],
            'units': coll.ddata[kknots[0]]['units'],
            'dim': coll.ddata[kknots[0]]['dim'],
            'quant': coll.ddata[kknots[0]]['quant'],
            'name': coll.ddata[kknots[0]]['name'],
            'ref': (keybsr,),
        },
        kbscz: {
            'data': bs_cents[1, :],
            'units': coll.ddata[kknots[1]]['units'],
            'dim': coll.ddata[kknots[1]]['dim'],
            'quant': coll.ddata[kknots[1]]['quant'],
            'name': coll.ddata[kknots[1]]['name'],
            'ref': (keybsr,),
        },
    }

    dobj = {
        'bsplines': {
            keybs: {
                'deg': deg,
                'mesh': keym,
                'ref': (keybsr,),
                'ref-bs': (keybsr,),
                'apex': (kbscr, kbscz),
                'shape': (clas.nbs,),
                'crop': False,
                'class': clas,
            }
        },
    }

    return dref, ddata, dobj


# ##############################################################
# ##############################################################
#                   Utils
# ##############################################################


def get_bs_class(
    nd=None,
    mtype=None,
    deg=None,
    knots0=None,
    knots1=None,
    shapebs=None,
    indices=None,
):

    if nd == '1d':
        clas = _class02_bsplines_1d.get_bs_class(
            deg=deg,
            knots=knots0,
        )

    elif mtype == 'rect':
        clas = _class02_bsplines_rect.get_bs_class(
            deg=deg,
            knots0=knots0,
            knots1=knots1,
            shapebs=shapebs,
        )

    else:
        clas = _class02_bsplines_tri.get_bs_class(
            deg=deg,
            knots0=knots0,
            knots1=knots1,
            indices=indices,
        )

    return clas


def _get_profiles2d(coll=None):

    dout = {}
    wm = coll._which_mesh
    wbs = coll._which_bsplines
    for k0, v0 in coll.ddata.items():

        if v0.get(wbs) is None or v0.get(wbs) == '':
            continue

        lbs = []
        for k1 in v0[wbs]:
            km = coll.dobj[wbs][k1][wm]
            subkm = coll.dobj[wm][km]['submesh']
            if subkm is None:
                if coll.dobj[wm][km]['nd'] == '2d':
                    lbs.append(k1)
            else:
                if coll.dobj[wm][subkm]['nd'] == '2d':
                    lbs.append(k1)

        if len(lbs) == 1:
            dout[k0] = lbs[0]

    return dout


# ##################################################################
# ##################################################################
#                       Extract
# ##################################################################


def extract(coll=None, coll2=None, vectors=None):

    wm = coll._which_mesh
    wbs = coll._which_bsplines

    # -------------
    # find bsplines and meshes

    lbs, lmesh, ldata = _extract_bsplines_mesh(
        coll=coll,
        coll2=coll2,
    )

    # ---------------
    # trivial

    if len(lmesh) == 0:
        return coll2

    # ---------------
    # extract data

    dmesh = _extract_refdata_from_obj(coll=coll, which=wm, keys=lmesh)
    dbs = _extract_refdata_from_obj(coll=coll, which=wbs, keys=lbs)

    # get all keys for mesh and bsplines
    kmesh = list(itt.chain.from_iterable([
        v0['ref'] + v0['data'] for v0 in dmesh.values()
    ]))
    kbs = list(itt.chain.from_iterable([
        v0['ref'] + v0['data'] for v0 in dbs.values()
    ]))

    # get all ref and data
    lref2, ldata2 = ds._class1_compute._extract_dataref(
        coll=coll,
        keys=list(set(kmesh + kbs + ldata)),
        vectors=vectors
    )

    # ---------------
    # add mesh

    coll2._dobj[wm] = {
        k0: copy.deepcopy(coll.dobj[wm][k0])
        for k0 in lmesh
    }

    # ----------------
    # add bsplines

    coll2._dobj[wbs] = {}
    for k0 in lbs:
        din = {
            k1: copy.deepcopy(v1)
            for k1, v1 in coll.dobj[wbs][k0].items()
            if k1 != 'class'
        }

        # clas has to be treated sepearately (cannot be copied)
        km = coll.dobj[wbs][k0][wm]
        nd = str(coll.dobj[wm][km]['nd'])
        mtype = str(coll.dobj[wm][km]['type'])
        deg = int(coll.dobj[wbs][k0]['deg'])
        kn = coll.dobj[wm][km]['knots']
        knots0 = np.copy(coll.ddata[kn[0]]['data'])
        knots1 = np.copy(coll.ddata[kn[1]]['data']) if nd == '2d' else None
        shapebs = tuple(coll.dobj[wbs][k0]['shape'])
        if mtype == 'tri':
            kind = coll2.dobj[wm][km]['ind']
            indices = np.copy(coll.ddata[kind]['data'])
        else:
            indices = None

        din['class'] = get_bs_class(
            nd=nd,
            mtype=mtype,
            deg=deg,
            knots0=knots0,
            knots1=knots1,
            shapebs=shapebs,
            indices=indices,
        )

        coll2._dobj[wbs][k0] = din

    # --------------------
    # extract

    coll2 = ds._class1_compute._extract_instance(
        coll=coll,
        lref=lref2,
        ldata=ldata2,
        coll2=coll2,
    )

    return coll2


def _extract_bsplines_mesh(coll=None, coll2=None):

    # -------------
    # find bsplines

    # bs from data['bsplines']
    wbs = coll._which_bsplines
    lbs = list(set(itt.chain.from_iterable([
        v0[wbs] for v0 in coll2.ddata.values()
        if isinstance(v0.get(wbs), tuple)
    ])))

    # bs from ref
    lbs2 = []
    if wbs in coll.dobj.keys():
        for k0, v0 in coll.dobj[wbs].items():
            for k1 in coll2.dref.keys():
                if k1 in v0['ref'] + v0['ref-bs']:
                    lbs2.append(k0)
            for k1 in coll2.ddata.keys():
                if k1 in v0['apex']:
                    lbs2.append(k0)

    lbs = list(set(lbs + lbs2))

    # -----------
    # find meshes

    wm = coll._which_mesh
    lmesh = list({coll.dobj[wbs][k0][wm] for k0 in lbs})

    # -----------
    # get submesh

    ldata = list(set(itt.chain.from_iterable([
        coll.dobj[wm][k0]['subkey'] for k0 in lmesh
        if coll.dobj[wm][k0]['subkey'] is not None
    ])))

    lbs2 = list(itt.chain.from_iterable([coll.ddata[k0][wbs] for k0 in ldata]))
    lmesh2 = list({coll.dobj[wbs][k0][wm] for k0 in lbs2})

    return sorted(set(lbs + lbs2)), sorted(set(lmesh + lmesh2)), ldata


def _extract_refdata_from_obj(coll=None, which=None, keys=None):

    dout = {}
    for k0 in keys:
        dout[k0] = {'ref': [], 'data': []}
        _extract_refdata_from_dict(
            dout=dout[k0],
            din=coll.dobj[which][k0],
            lref=list(coll.dref.keys()),
            ldata=list(coll.ddata.keys()),
        )

    return dout


def _extract_refdata_from_dict(
    dout=None,
    din=None,
    lref=None,
    ldata=None,
):

    for k1, v1 in din.items():

        if isinstance(v1, str):
            if v1 in lref:
                dout['ref'].append(v1)
            elif v1 in ldata:
                dout['data'].append(v1)

        elif isinstance(v1, (list, tuple)):
            for v2 in v1:
                if not isinstance(v2, str):
                    continue
                if v2 in lref:
                    dout['ref'].append(v2)
                elif v2 in ldata:
                    dout['data'].append(v2)

        elif isinstance(v1, dict):
            _extract_refdata_from_dict(
                dout=dout,
                din=v1,
                lref=lref,
                ldata=ldata,
            )


# ##################################################################
# ##################################################################
#                           Mesh2D - polar - bsplines
# ##################################################################


# def _mesh2Dpolar_bsplines(
    # coll=None,
    # keym=None,
    # keybs=None,
    # angle=None,
    # deg=None,
# ):

    # # ---------------
    # # create bsplines

    # kknots = coll.dobj[coll._which_mesh][keym]['knots']
    # knotsr = coll.ddata[kknots[0]]['data']
    # if len(kknots) == 2:
        # angle = coll.ddata[kknots[1]]['data']

    # clas = _class02_bsplines_polar.get_bs2d_func(
        # deg=deg,
        # knotsr=knotsr,
        # angle=angle,
        # coll=coll,
    # )

    # keybsnr = f'{keybs}-nr'
    # keybsn = f'{keybs}-nbs'
    # keybsapr = f'{keybs}-apr'

    # # ------------
    # # refs

    # if clas.knotsa is None:
        # ref = (keybsnr,)
        # apex = (keybsapr,)
    # elif len(clas.shapebs) == 2:
        # keybsna = f'{keybs}-na'
        # keybsapa = f'{keybs}-apa'
        # ref = (keybsnr, keybsna)
        # apex = (keybsapr, keybsapa)
    # else:
        # ref = (keybsn,)
        # apex = (keybsapr,)

        # # check angle vs angle2d
        # mesh = coll._which_mesh
        # angle2d = coll.dobj[mesh][keym]['angle2d']
        # if angle2d is None:
            # msg = (
                # "Poloidal bsplines require mesh with angle2d!\n"
                # f"\t- self.dobj['{mesh}']['{keym}']['angle2d'] = {angle2d}"
            # )
            # raise Exception(msg)

    # # bs_cents = clas._get_bs_cents()

    # # ----------------
    # # format into dict

    # # dref
    # dref = {
        # # bs index
        # keybsnr: {'size': clas.nbs_r},
        # keybsn: {'size': clas.nbs},
    # }
    # if len(clas.shapebs) == 2:
        # dref[keybsna] = {'size': clas.nbs_a_per_r[0]}

    # # ddata
    # ddata = {
        # keybsapr: {
            # 'data': clas.apex_per_bs_r,
            # 'units': '',
            # 'dim': '',
            # 'quant': '',
            # 'name': '',
            # 'ref': (keybsnr,),
        # },
    # }
    # if len(clas.shapebs) == 2:
        # ddata[keybsapa] = {
            # 'data': clas.apex_per_bs_a[0],
            # 'units': 'rad',
            # 'dim': 'angle',
            # 'quant': '',
            # 'name': '',
            # 'ref': (keybsna,),
        # }

    # # dobj
    # dobj = {
        # 'bsplines': {
            # keybs: {
                # 'deg': deg,
                # 'mesh': keym,
                # 'ref': ref,
                # 'ref-bs': (keybsn,),
                # 'apex': apex,
                # 'shape': clas.shapebs,
                # 'class': clas,
                # 'crop': coll.dobj[coll._which_mesh][keym]['crop'],
            # }
        # },
    # }

    # return dref, ddata, dobj


# def _mesh2DPolar_bsplines_knotscents(
    # returnas=None,
    # return_knots=None,
    # return_cents=None,
    # ind=None,
    # deg=None,
    # # resources
    # clas=None,
    # rknots=None,
    # aknots=None,
    # rcents=None,
    # acents=None,
# ):

    # # -------------
    # # check inputs

    # return_knots = ds._generic_check._check_var(
        # return_knots, 'return_knots',
        # types=bool,
        # default=True,
    # )
    # return_cents = ds._generic_check._check_var(
        # return_cents, 'return_cents',
        # types=bool,
        # default=True,
    # )
    # if return_knots is False and return_cents is False:
        # return

    # # -------------
    # # compute

    # if return_knots is True:

        # knots_per_bs_r = _utils_bsplines._get_knots_per_bs(
            # rknots, deg=deg, returnas=returnas,
        # )
        # knots_per_bs_Z = _utils_bsplines._get_knots_per_bs(
            # Zknots, deg=deg, returnas=returnas,
        # )
        # if ind is not None:
            # knots_per_bs_R = knots_per_bs_R[:, ind[0]]
            # knots_per_bs_Z = knots_per_bs_Z[:, ind[1]]

        # nknots = knots_per_bs_R.shape[0]
        # knots_per_bs_R = np.tile(knots_per_bs_R, (nknots, 1))
        # knots_per_bs_Z = np.repeat(knots_per_bs_Z, nknots, axis=0)

    # if return_cents is True:

        # cents_per_bs_R = _utils_bsplines._get_cents_per_bs(
            # Rcents, deg=deg, returnas=returnas,
        # )
        # cents_per_bs_Z = _utils_bsplines._get_cents_per_bs(
            # Zcents, deg=deg, returnas=returnas,
        # )
        # if ind is not None:
            # cents_per_bs_R = cents_per_bs_R[:, ind[0]]
            # cents_per_bs_Z = cents_per_bs_Z[:, ind[1]]

        # ncents = cents_per_bs_R.shape[0]
        # cents_per_bs_R = np.tile(cents_per_bs_R, (ncents, 1))
        # cents_per_bs_Z = np.repeat(cents_per_bs_Z, ncents, axis=0)

    # # -------------
    # # return

    # if return_knots is True and return_cents is True:
        # out = (
            # (knots_per_bs_R, knots_per_bs_Z), (cents_per_bs_R, cents_per_bs_Z)
        # )
    # elif return_knots is True:
        # out = (knots_per_bs_R, knots_per_bs_Z)
    # else:
        # out = (cents_per_bs_R, cents_per_bs_Z)
    # return out


# #############################################################################
# #############################################################################
#                           Mesh2DRect - interp utility
# #############################################################################


# def _get_keyingroup_ddata(
    # dd=None, dd_name='data',
    # key=None, monot=None,
    # msgstr=None, raise_=False,
# ):
    # """ Return the unique data key matching key

    # Here, key can be interpreted as name / source / units / quant...
    # All are tested using select() and a unique match is returned
    # If not unique match an error message is either returned or raised

    # """

    # # ------------------------
    # # Trivial case: key is actually a ddata key

    # if key in dd.keys():
        # return key, None

    # # ------------------------
    # # Non-trivial: check for a unique match on other params

    # dind = _select(
        # dd=dd, dd_name=dd_name,
        # dim=key, quant=key, name=key, units=key, source=key,
        # monot=monot,
        # log='raw',
        # returnas=bool,
    # )
    # ind = np.array([ind for kk, ind in dind.items()])

    # # Any perfect match ?
    # nind = np.sum(ind, axis=1)
    # sol = (nind == 1).nonzero()[0]
    # key_out, msg = None, None
    # if sol.size > 0:
        # if np.unique(sol).size == 1:
            # indkey = ind[sol[0], :].nonzero()[0]
            # key_out = list(dd.keys())[indkey]
        # else:
            # lstr = "[dim, quant, name, units, source]"
            # msg = "Several possible matches in {} for {}".format(lstr, key)
    # else:
        # lstr = "[dim, quant, name, units, source]"
        # msg = "No match in {} for {}".format(lstr, key)

    # # Complement error msg and optionally raise
    # if msg is not None:
        # lk = ['dim', 'quant', 'name', 'units', 'source']
        # dk = {
            # kk: (
                # dind[kk].sum(),
                # sorted(set([vv[kk] for vv in dd.values()]))
            # ) for kk in lk
        # }
        # msg += (
            # "\n\nRequested {} could not be identified!\n".format(msgstr)
            # + "Please provide a valid (unique) key/name/dim/quant/units:\n\n"
            # + '\n'.join([
                # '\t- {} ({} matches): {}'.format(kk, dk[kk][0], dk[kk][1])
                # for kk in lk
            # ])
            # + "\nProvided:\n\t'{}'".format(key)
        # )
        # if raise_:
            # raise Exception(msg)
    # return key_out, msg

"""
def _get_possible_ref12d(
    dd=None,
    key=None, ref1d=None, ref2d=None,
    group1d='radius',
    group2d='mesh2d',
):

    # Get relevant lists
    kq, msg = _get_keyingroup_ddata(
        dd=dd,
        key=key, group=group2d, msgstr='quant', raise_=False,
    )

    if kq is not None:
        # The desired quantity is already 2d
        k1d, k2d = None, None

    else:
        # Check if the desired quantity is 1d
        kq, msg = _get_keyingroup_ddata(
            dd=dd,
            key=key, group=group1d,
            msgstr='quant', raise_=True,
        )

        # Get dict of possible {ref1d: lref2d}
        ref = [rr for rr in dd[kq]['ref'] if dd[rr]['group'] == (group1d,)][0]
        lref1d = [
            k0 for k0, v0 in dd.items()
            if ref in v0['ref'] and v0['monot'][v0['ref'].index(ref)] is True
        ]

        # Get matching ref2d with same quant and good group
        lquant = list(set([dd[kk]['quant'] for kk in lref1d]))
        dref2d = {
            k0: [
                kk for kk in _select(
                    dd=dd, quant=dd[k0]['quant'],
                    log='all', returnas=str,
                )
                if group2d in dd[kk]['group']
                and not isinstance(dd[kk]['data'], dict)
            ]
            for k0 in lref1d
        }
        dref2d = {k0: v0 for k0, v0 in dref2d.items() if len(v0) > 0}

        if len(dref2d) == 0:
            msg = (
                "No match for (ref1d, ref2d) for ddata['{}']".format(kq)
            )
            raise Exception(msg)

        # check ref1d
        if ref1d is None:
            if ref2d is not None:
                lk = [k0 for k0, v0 in dref2d.items() if ref2d in v0]
                if len(lk) == 0:
                    msg = (
                        "\nNon-valid interpolation intermediate\n"
                        + "\t- provided:\n"
                        + "\t\t- ref1d = {}, ref2d = {}\n".format(ref1d, ref2d)
                        + "\t- valid:\n{}".format(
                            '\n'.join([
                                '\t\t- ref1d = {}  =>  ref2d in {}'.format(
                                    k0, v0
                                )
                                for k0, v0 in dref2d.items()
                            ])
                        )
                    )
                    raise Exception(msg)
                if kq in lk:
                    ref1d = kq
                else:
                    ref1d = lk[0]
            else:
                if kq in dref2d.keys():
                    ref1d = kq
                else:
                    ref1d = list(dref2d.keys())[0]
        else:
            ref1d, msg = _get_keyingroup_ddata(
                dd=dd,
                key=ref1d, group=group1d,
                msgstr='ref1d', raise_=False,
            )
        if ref1d not in dref2d.keys():
            msg = (
                "\nNon-valid interpolation intermediate\n"
                + "\t- provided:\n"
                + "\t\t- ref1d = {}, ref2d = {}\n".format(ref1d, ref2d)
                + "\t- valid:\n{}".format(
                    '\n'.join([
                        '\t\t- ref1d = {}  =>  ref2d in {}'.format(
                            k0, v0
                        )
                        for k0, v0 in dref2d.items()
                    ])
                )
            )
            raise Exception(msg)

        # check ref2d
        if ref2d is None:
            ref2d = dref2d[ref1d][0]
        else:
            ref2d, msg = _get_keyingroup_ddata(
                dd=dd,
                key=ref2d, group=group2d,
                msgstr='ref2d', raise_=False,
            )
        if ref2d not in dref2d[ref1d]:
            msg = (
                "\nNon-valid interpolation intermediate\n"
                + "\t- provided:\n"
                + "\t\t- ref1d = {}, ref2d = {}\n".format(ref1d, ref2d)
                + "\t- valid:\n{}".format(
                    '\n'.join([
                        '\t\t- ref1d = {}  =>  ref2d in {}'.format(
                            k0, v0
                        )
                        for k0, v0 in dref2d.items()
                    ])
                )
            )
            raise Exception(msg)

    return (
        kq, ref1d, ref2d,
        coll=None,
        key=None,
        R=None,
        Z=None,
        indbs=None,
        indt=None,
        grid=None,
        details=None,
        reshape=None,
        res=None,
        crop=None,
        nan0=None,
        imshow=None,
        return_params=None,
    )
):

    # ---------------
    # check inputs

    # TBD
    pass

    # ---------------
    # post-treatment

    if nan0 is True:
        val[val == 0] = np.nan

    # ------
    # return

    if return_params is True:
        return val, dparams
    else:
        return val
"""



"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import itertools as itt
import warnings


# Standard
import numpy as np
import matplotlib.pyplot as plt


# specific
from . import test_input


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module():
    pass

def teardown_module():
    pass


#######################################################
#
#     object mesh2D
#
#######################################################


"""
class Test01_BSplines():

    @classmethod
    def setup_class(cls):
        pass

    def setup_method(self):
        bsplines = tfd.Collection()

        # add rect mesh
        _add_rect_uniform(bsplines)
        _add_rect_variable(bsplines)
        _add_rect_variable_crop(bsplines)

        # add tri mesh
        _add_tri_ntri1(bsplines)
        _add_tri_ntri2(bsplines)

        # add bsplines
        _add_bsplines(bsplines)

        # add polar mesh
        _add_polar1(bsplines)
        _add_polar2(bsplines)

        # add bsplines for polar meshes
        _add_bsplines(bsplines, kind=['polar'])

        # Add polar with variable poloidal discretization
        _add_polar2(bsplines, key='m7')
        _add_bsplines(
            bsplines,
            key=['m7'],
            angle=np.pi*np.r_[-3./4., -1/4, 0, 1/4, 3/4],
        )

        # store
        self.obj = bsplines
        self.lm = list(bsplines.dobj['mesh'].keys())
        self.lbs = list(bsplines.dobj['bsplines'].keys())

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test04_select_bsplines(self):

        lind0 = [None, ([0, 2], [0, 4]), [0, 2, 4], ([0, 2, 4], [0, 2, 3])]
        lind1 = [None, [1], 1, [0, 1]]
        for ii, k0 in enumerate(self.lbs):

            km = self.obj.dobj['bsplines'][k0]['mesh']
            if len(self.obj.dobj['bsplines'][k0]['shape']) == 2:
                lind = lind0
            else:
                lind = lind1

            if self.obj.dobj['mesh'][km]['type'] == 'polar':
                return_cents = False
                return_knots = False
            else:
                return_cents = None if ii == 1 else bool(ii%3)
                return_knots = None if ii == 2 else bool(ii%2)

            out = self.obj.select_bsplines(
                key=k0,
                ind=lind[ii%len(lind)],
                returnas='ind' if ii%3 == 0 else 'data',
                return_cents=return_cents,
                return_knots=return_knots,
            )

    def test08_interpolate_profile2d_details_vs_sum(self):

        x = np.linspace(2.2, 2.8, 5)
        y = np.linspace(-0.5, 0.5, 5)
        x = np.tile(x, (y.size, 1))
        y = np.tile(y, (x.shape[1], 1)).T

        for ii, k0 in enumerate(self.lbs):

            keym = self.obj.dobj['bsplines'][k0]['mesh']
            mtype = self.obj.dobj['mesh'][keym]['type']

            val, t, ref = self.obj.interpolate_profile2d(
                key=k0,
                R=x,
                Z=y,
                coefs=None,
                indbs=None,
                indt=None,
                grid=False,
                details=True,
                reshape=True,
                res=None,
                crop=None,
                nan0=ii % 2 == 0,
                imshow=False,
                return_params=False,
            )

            crop = self.obj.dobj['bsplines'][k0].get('crop', False)
            nbs = np.prod(self.obj.dobj['bsplines'][k0]['shape'])
            if isinstance(crop, str):
                nbs = self.obj.ddata[crop]['data'].sum()

            vshap0 = tuple(np.r_[x.shape, nbs])
            if mtype == 'polar':
                # radius2d can be time-dependent => additional dimension
                vshap = val.shape[-len(vshap0):]
            else:
                vshap = val.shape
            assert vshap == vshap0, val.shape

            val_sum, t, ref = self.obj.interpolate_profile2d(
                key=k0,
                R=x,
                Z=y,
                coefs=None,
                indbs=None,
                indt=None,
                grid=False,
                details=False,
                reshape=True,
                res=None,
                crop=None,
                nan0=False,
                val_out=0.,
                imshow=False,
                return_params=False,
            )

            if mtype == 'polar':
                # radius2d can be time-dependent => additional dimension
                vshap_sum = val_sum.shape[-len(x.shape):]
            else:
                vshap_sum = val_sum.shape
            assert vshap_sum == x.shape, val_sum.shape
            assert (val.ndim == x.ndim + 2) == (val_sum.ndim == x.ndim + 1), [val.shape, val_sum.shape]

            indok = np.isfinite(val_sum)
            indok[indok] = val_sum[indok] != 0

            # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            # Does not work for rect mesh
            # because of knots padding used in func_details
            # Due to scpinterp._bspl.evaluate_spline()...
            if mtype in ['tri', 'polar']:   # To be debugged
                assert np.allclose(
                    val_sum[indok],
                    np.nansum(val, axis=-1)[indok],
                    equal_nan=True,
                )
            # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    def test10_plot_bsplines(self):

        li0 = [None, ([1, 2], [2, 1]), (1, 1), [1, 2, 4]]
        li1 = [None, [1, 2], (1, 1), [1, 2, 4]]
        for ii, k0 in enumerate(self.lbs):

            km = self.obj.dobj['bsplines'][k0]['mesh']
            if len(self.obj.dobj['mesh'][km]['shape-c']) == 2:
                li = li0
            else:
                li = li1

            if self.obj.dobj['mesh'][km]['type'] == 'polar':
                plot_mesh = False
            else:
                plot_mesh = True

            dax = self.obj.plot_bsplines(
                key=k0,
                indbs=li[ii%len(li)],
                knots=bool(ii%3),
                cents=bool(ii%2),
                res=0.05,
                plot_mesh=plot_mesh,
            )
            plt.close('all')

    def test12_add_bsplines_operator(self):
        lkey = ['m0-bs0', 'm1-bs1', 'm2-bs2']
        lop = ['D0N1', 'D0N2', 'D1N2', 'D2N2']
        lgeom = ['linear', 'toroidal']
        lcrop = [False, True]

        dfail = {}
        for ii, k0 in enumerate(self.lbs):

            km = self.obj.dobj['bsplines'][k0]['mesh']
            if self.obj.dobj['mesh'][km]['type'] == 'tri':
                continue
            elif self.obj.dobj['mesh'][km]['type'] == 'polar':
                continue

            for comb in itt.product(lop, lgeom, lcrop):
                deg = self.obj.dobj['bsplines'][k0]['deg']

                if deg == 3 and comb[0] in ['D0N1', 'D0N2', 'D1N2', 'D2N2']:
                    continue

                # only test exact operators
                if int(comb[0][1]) > deg:
                    # except deg = 0 D1N2
                    if deg == 0 and comb[0] == 'D1N2':
                        pass
                    else:
                        continue
                try:
                    self.obj.add_bsplines_operator(
                        key=k0,
                        operator=comb[0],
                        geometry=comb[1],
                        crop=comb[2],
                    )
                except Exception as err:
                    dfail[k0] = (
                        f"key {k0}, op '{comb[0]}', geom '{comb[1]}': "
                        + str(err)
                    )

        # Raise error if any fail
        if len(dfail) > 0:
            lstr = [f'\t- {k0}: {v0}' for k0, v0 in dfail.items()]
            msg = (
                "The following operators failed:\n"
                + "\n".join(lstr)
            )
            raise Exception(msg)
"""



# -*- coding: utf-8 -*-


# Built-in
import copy
import itertools as itt


# Common
import datastock as ds

# local


# ###########################################################
# ###########################################################
#               Main routine
# ###########################################################


def interpolate_all_bsplines(
    coll=None,
    keys=None,
    # sampling
    dres=None,
    submesh=None,
    dunique_mesh_2d=None,
    # parameters
    val_out=None,
    nan0=None,
):

    # ----------
    # check

    keys, coll2, dres, submesh = _check(
        coll=coll,
        keys=keys,
        # sampling
        dres=dres,
        submesh=submesh,
    )

    # ---------
    # unique

    if dunique_mesh_2d is not None:
        for k0, v0 in dunique_mesh_2d.items():
            kref = f"{v0['key']}_n"
            coll2.add_ref(key=kref, size=v0['data'].size)
            coll2.add_data(ref=kref, **v0)
            dunique_mesh_2d[k0]['ref'] = kref
        dout = dunique_mesh_2d

    # ----------------
    # interpolate loop

    dbs = {}
    wm = coll._which_mesh
    wbs = coll._which_bsplines
    for ii, (k0, v0) in enumerate(dres.items()):

        # get 2d mesh
        if dunique_mesh_2d is None:
            if v0.get('x0') is None:
                dout = coll2.get_sample_mesh(
                    key=k0,
                    res=v0['res'],
                    mode=v0['mode'],
                    grid=False,
                    # store
                    store=True,
                    kx0=None,
                    kx1=None,
                )
            else:
                x0str = f'{k0}_x0_temp'
                dout = {'x0': {'key': x0str, 'ref': f'{x0str}_n'}}
                if v0.get('x1') is not None:
                    x1str = f'{k0}_x1_temp'
                    dout['x1'] = {'key': x1str, 'ref': f'{x1str}_n'}
                    for k1, v1 in dout.items():
                        coll2.add_ref(key=v1['ref'], size=v0[k1].size)
                        coll2.add_data(
                            key=v1['key'],
                            data=v0[k1],
                            ref=v1['ref'],
                        )

        # compute
        for bs in v0[wbs]:

            if bs not in dbs.keys():
                dbs[bs] = {'ref': tuple([v1['ref'] for v1 in dout.values()])}

            for key in keys:

                if bs not in coll2.ddata[key][wbs]:
                    continue

                # submesh => ref_com
                kms = coll2.dobj[wbs][bs][wm]
                subkey = coll2.dobj[wm][kms]['subkey']
                if submesh is True and subkey is not None:
                    refsub = coll2.ddata[subkey[0]]['ref']
                    ref = coll2.ddata[key]['ref']
                    if refsub[0] in ref:
                        ref_com = refsub[0]
                    elif refsub[-1] in ref:
                        ref_com = refsub[-1]
                    else:
                        ref_com = None
                else:
                    ref_com = None

                # interpolate
                coll2 = coll2.interpolate(
                    keys=key,
                    ref_key=bs,
                    x0=dout['x0']['key'],
                    x1=dout.get('x1', {}).get('key'),
                    submesh=submesh,
                    ref_com=ref_com,
                    grid=True,
                    details=False,
                    # parameters
                    val_out=val_out,
                    nan0=nan0,
                    # return vs store
                    returnas=object,
                    return_params=False,
                    store=True,
                    inplace=True,
                )

                # remove original data
                if key in coll2.ddata.keys():
                    coll2.remove_data(key)

                # rename new data
                keynew = f'{key}_interp'
                coll2.add_data(
                    key=key,
                    **{k1: v1 for k1, v1 in coll2.ddata[keynew].items()},
                )

                # remove keynew
                coll2.remove_data(keynew)

    # -----------
    # clean-up

    lbs = list(set(itt.chain.from_iterable([v0[wbs] for v0 in dres.values()])))
    if wbs in coll2.dobj.keys():
        coll2.remove_bsplines(key=lbs, propagate=True)

    if wm in coll2.dobj.keys():
        coll2.remove_mesh(key=list(dres.keys()), propagate=True)

    return coll2, dbs


# ###########################################################
# ###########################################################
#               check
# ###########################################################


def _check(
    coll=None,
    keys=None,
    # sampling
    knots0=None,
    knots1=None,
    dres=None,
    submesh=None,
):

    # ----
    # key

    wm = coll._which_mesh
    wbs = coll._which_bsplines

    lok = [
        k0 for k0, v0 in coll.ddata.items()
        if isinstance(v0[wbs], tuple)
    ]

    if isinstance(keys, str):
        keys = [keys]
    keys = ds._generic_check._check_var_iter(
        keys, 'keys',
        types=(list, tuple),
        types_iter=str,
        allowed=lok,
    )

    lbs = list(itt.chain.from_iterable([coll.ddata[k0][wbs] for k0 in keys]))

    # --------------
    # rbs vs submesh

    submesh = ds._generic_check._check_var(
        submesh, 'submesh',
        types=bool,
        default=True,
    )

    dbs = {}
    for ii, bs in enumerate(lbs):
        km = coll.dobj[wbs][bs][wm]
        if submesh and coll.dobj[wm][km]['submesh'] is not None:
            km = coll.dobj[wm][km]['submesh']
        dbs[bs] = km

    # --------------
    # derive mesh

    lm = list(set([dbs[k0] for k0 in lbs]))

    # --------------
    # dres

    if dres is None:
        dres = {k0: {'res': None, 'mode': 'abs'} for k0 in lm}

    if isinstance(dres, (int, float)):
        dres = {k0: {'res': dres, 'mode': 'abs'} for k0 in lm}

    # safety check
    c0 = (
        isinstance(dres, dict)
        and all([kk in lm for kk in dres.keys()])
    )
    if not c0:
        msg = (
            "Arg dres must be a dict with, for each mesh\n"
            "\t- {'res': float, 'mode': str}\n"
            f"\nFor the following keys ({wm}): {lm}\n"
            f"Provided:\n{dres}\n"
        )
        raise Exception(msg)

    # loop
    for k0 in lm:
        if dres.get(k0) is None:
            dres[k0] = {'res': None, 'mode': 'abs'}
        elif isinstance(dres[k0], dict):
            dres[k0] = {
                'res': dres[k0].get('res'),
                'mode': dres[k0].get('mode', 'abs'),
                'x0': dres[k0].get('x0'),
                'x1': dres[k0].get('x1'),
            }
        elif isinstance(dres[k0], (float, int)):
            dres[k0] = {
                'res': dres[k0],
                'mode': 'abs',
            }
        else:
            msg = (
                "Arg dres must be a dict with, for each mesh\n"
                "\t- {'res': float, 'mode': str}\n"
                f"Provided:\n{dres}"
            )
            raise Exception(msg)

        # add bsplines
        dres[k0][wbs] = [k1 for k1 in lbs if dbs[k1] == k0]

    # ----------
    # coll2

    coll2 = coll.extract(keys=keys, vectors=True)

    return keys, coll2, dres, submesh



# -*- coding: utf-8 -*-


# Built-in
import itertools as itt


# Common
import numpy as np
import scipy.interpolate as scpinterp


# specific
from . import _utils_bsplines
from . import _class02_bsplines_operators_rect


# ################################################################
# ################################################################
#                       BivariateSplineRect - scipy subclass
# ################################################################


class BivariateSplineRect(scpinterp.BivariateSpline):
    """ Subclass for tofu

    Defined from knots (unique) and deg
    coefs set to 1 by default

    Used self.set_coefs() to update
    """

    def __init__(self, knots0=None, knots1=None, deg=None, shapebs=None):

        assert np.allclose(np.unique(knots0), knots0)
        assert np.allclose(np.unique(knots1), knots1)
        assert deg in [0, 1, 2, 3]

        # get knots pr bs
        self._get_knots_per_bs_for_basis_elements(
            knots0=knots0,
            knots1=knots1,
            deg=deg,
        )

        # full knots with multiplicity
        knots0, nbs0 = _utils_bsplines._get_knots_per_bs(
            knots0, deg=deg, returnas='data', return_unique=True,
        )
        knots1, nbs1 = _utils_bsplines._get_knots_per_bs(
            knots1, deg=deg, returnas='data', return_unique=True,
        )

        coefs = np.ones((nbs0*nbs1,), dtype=float)

        self.__nbs = (nbs0, nbs1)
        self.tck = [knots0, knots1, coefs]
        self.degrees = [deg, deg]

        # shapebs
        self.shapebs = shapebs

    def _get_knots_per_bs_for_basis_elements(
        self,
        knots0=None,
        knots1=None,
        deg=None,
    ):

        # added for details
        knots_per_bs_x0 = _utils_bsplines._get_knots_per_bs(
            knots0, deg=deg, returnas='data',
        )
        knots_per_bs_x1 = _utils_bsplines._get_knots_per_bs(
            knots1, deg=deg, returnas='data',
        )

        self.knots_per_bs_x0 = knots_per_bs_x0
        self.knots_per_bs_x1 = knots_per_bs_x1

        if deg == 0:
            pass
        else:
            knots_per_bs_x0 = np.concatenate(
                (
                    np.tile(knots_per_bs_x0[0, :] - 1, (deg, 1)),
                    knots_per_bs_x0,
                    np.tile(knots_per_bs_x0[-1, :] + 1, (deg, 1)),
                ),
                axis=0,
            )
            knots_per_bs_x1 = np.concatenate(
                (
                    np.tile(knots_per_bs_x1[0, :] - 1, (deg, 1)),
                    knots_per_bs_x1,
                    np.tile(knots_per_bs_x1[-1, :] + 1, (deg, 1)),
                ),
                axis=0,
            )

        self.knots_per_bs_x0_pad = np.asfortranarray(knots_per_bs_x0)
        self.knots_per_bs_x1_pad = np.asfortranarray(knots_per_bs_x1)

    def set_coefs(
        self,
        coefs=None,
        cropbs_neg_flat=None,
    ):

        if coefs.shape == self.shapebs:
            self.tck[2][...] = coefs.ravel()
        elif coefs.shape == (self.nbs,):
            self.tck[2][...] = coefs
        else:
            msg = f"Wrong coefs shape!\nProvided: {coefs.shape}"
            raise Exception(msg)

        # ------------
        # crop and set

        if cropbs_neg_flat is not None:
            self.tck[2][cropbs_neg_flat] = 0.

    def __call__(
        self,
        # interp points
        x0=None,
        x1=None,
        # coefs
        coefs=None,
        axis=None,
        # crop
        crop=None,
        cropbs=None,
        # options
        val_out=None,
        # slicing
        sli_c=None,
        sli_x=None,
        sli_v=None,
        sli_o=None,
        indokx0=None,
        shape_v=None,
        shape_o=None,
        dref_com=None,
        # for compatibility (unused)
        **kwdargs,
    ):

        # prepare
        val = np.zeros(shape_v, dtype=float)
        cropbs_neg_flat = (~cropbs).ravel() if crop else None

        # interpolate
        for ind in itt.product(*[range(aa) for aa in shape_o]):

            # slices
            slic = sli_c(
                ind,
                axis=axis,
                ddim=coefs.ndim,
            )

            slix = sli_x(
                ind,
                indokx0=indokx0,
                ix=dref_com['ix'],
                iother=dref_com['iother'],
            )

            sliv = sli_v(
                ind,
                indokx0=indokx0,
                ddim=coefs.ndim,
                axis=axis,
                ix=dref_com['ix'],
                iother=dref_com['iother'],
            )

            self.set_coefs(
                coefs=coefs[slic],
                cropbs_neg_flat=cropbs_neg_flat,
            )

            # can be called on any shape of x0, x1?
            val[sliv] = super().__call__(x0[slix], x1[slix], grid=False)

        # clean out-of-mesh
        if dref_com['ix'] is None and val_out is not False:
            indout = (
                (x0 < self.tck[0][0]) | (x0 > self.tck[0][-1])
                | (x1 < self.tck[1][0]) | (x1 > self.tck[1][-1])
            )
            slio = sli_o(indout)
            val[slio] = val_out

        return val

    def ev_details(
        self,
        x0=None,
        x1=None,
        indbs_tf=None,
        crop=None,
        cropbs=None,
        # for compatibility (unused)
        **kwdargs,
    ):
        """
        indbs_tf = (ar0, ar1)
            tuple of 2 flat arrays of int (for R and Z)
        """

        # -----------
        # prepare

        deg = self.degrees[0]
        nbs = indbs_tf[0].size
        shape = x0.shape
        x0 = np.ascontiguousarray(x0.ravel(), dtype=float)
        x1 = np.ascontiguousarray(x1.ravel(), dtype=float)
        coef = np.zeros((deg + 4, 1), dtype=float)
        coef[deg] = 1.
        out1 = np.full((x0.size, 1), np.nan)

        # -----------
        # compute

        val = np.zeros(tuple(np.r_[x0.shape, nbs]))
        indtot = np.arange(0, nbs)

        iz_u = np.unique(indbs_tf[1])

        for iz in iz_u:

            scpinterp._bspl.evaluate_spline(
                self.knots_per_bs_x1_pad[:, iz],
                coef,
                self.degrees[1],
                x1,
                0,
                False,
                out1,
            )

            indok1 = ~np.isnan(out1)
            if not np.any(indok1):
                continue
            indok0 = np.copy(indok1)

            ind0 = indbs_tf[1] == iz
            i0 = indbs_tf[0][ind0]
            for ii, ii0 in enumerate(i0):

                msg = (
                    f'\t bspline {indtot[ind0][ii]} / {nbs}   ({x0.size} pts)'
                )
                print(msg, end='\r', flush=True)

                if ii > 0:
                    indok0[...] = indok1

                out0 = np.full((indok1.sum(), 1), np.nan)

                scpinterp._bspl.evaluate_spline(
                    self.knots_per_bs_x0_pad[:, ii0],
                    coef,
                    self.degrees[0],
                    x0[indok1[:, 0]],
                    0,
                    False,
                    out0,
                )

                ixok = ~np.isnan(out0)
                if not np.any(ixok):
                    continue

                indok0[indok1] = ixok[:, 0]
                val[indok0[:, 0], indtot[ind0][ii]] = (out0[ixok]*out1[indok0])

        if shape != x0.shape:
            val = np.reshape(val, tuple(np.r_[shape, -1]))

        return val

    def get_overlap(self):
        return _get_overlap(
            deg=self.degrees[0],
            knots0=self.knots_per_bs_x0,
            knots1=self.knots_per_bs_x1,
            shapebs=self.shapebs,
        )

    def get_operator(
        self,
        operator=None,
        geometry=None,
        cropbs_flat=None,
        # specific to deg = 0
        cropbs=None,
        centered=None,
        # to return gradR, gradZ, for D1N2 deg 0, for tomotok
        returnas_element=None,
    ):
        """ Get desired operator """
        return _class02_bsplines_operators_rect.get_mesh2dRect_operators(
            deg=self.degrees[0],
            operator=operator,
            geometry=geometry,
            knotsx_mult=self.tck[0],
            knotsy_mult=self.tck[1],
            knotsx_per_bs=self.knots_per_bs_x0,
            knotsy_per_bs=self.knots_per_bs_x1,
            overlap=self.get_overlap(),
            cropbs_flat=cropbs_flat,
            # specific to deg = 0
            cropbs=cropbs,
            centered=centered,
            # to return gradR, gradZ, for D1N2 deg 0, for tomotok
            returnas_element=returnas_element,
        )


# ################################################################
# ################################################################
#                       Mesh2DRect - bsplines - overlap
# ################################################################


def _get_overlap(
    deg=None,
    knots0=None,
    knots1=None,
    shapebs=None,
):
    # nb of overlapping, inc. itself in 1d
    nbs0, nbs1 = shapebs
    ind00 = np.tile(np.arange(0, nbs0), nbs1)
    ind10 = np.repeat(np.arange(0, nbs1), nbs0)

    # complete
    ntot = 2*deg + 1

    add0= np.tile(np.arange(-deg, deg+1), ntot)
    add1 = np.repeat(np.arange(-deg, deg+1), ntot)

    inter0 = ind00[None, :] + add0[:, None]
    inter1 = ind10[None, :] + add1[:, None]

    # purge
    inter = inter0 + inter1*nbs0
    indneg = (
        (inter0 < 0) | (inter0 >= nbs0) | (inter1 < 0) | (inter1 >= nbs1)
    )
    inter[indneg] = -1

    return inter


# #############################################################################
# #############################################################################
#                           Mesh2DRect - bsplines
# #############################################################################


def get_bs2d_x01(deg=None, knots0=None, knots1=None):

    # ----------------
    # get knots per bspline, nb of bsplines...

    knots_per_bs_x0 = _utils_bsplines._get_knots_per_bs(
        knots0, deg=deg, returnas='data',
    )
    knots_per_bs_x1 = _utils_bsplines._get_knots_per_bs(
        knots1, deg=deg, returnas='data',
    )
    # nbkbs = knots_per_bs_R.shape[0]
    shapebs = (knots_per_bs_x0.shape[1], knots_per_bs_x1.shape[1])

    # ----------------
    # get centers of bsplines

    x0bs_apex = _utils_bsplines._get_apex_per_bs(
        knots=knots0,
        knots_per_bs=knots_per_bs_x0,
        deg=deg
    )
    x1bs_apex = _utils_bsplines._get_apex_per_bs(
        knots=knots1,
        knots_per_bs=knots_per_bs_x1,
        deg=deg
    )
    return shapebs, x0bs_apex, x1bs_apex, knots_per_bs_x0, knots_per_bs_x1


def get_bs_class(
    deg=None,
    knots0=None,
    knots1=None,
    shapebs=None,
    # knots_per_bs_R=None,
    # knots_per_bs_Z=None,
):

    # ----------------
    # Define functions

    return BivariateSplineRect(
        knots0=knots0,
        knots1=knots1,
        deg=deg,
        shapebs=shapebs,
    )



# -*- coding: utf-8 -*-


# Built-in


# Common
import astropy.units as asunits
import datastock as ds


# specific


# #############################################################################
# #############################################################################
#                           get operators
# #############################################################################


def get_bsplines_operator(
    coll=None,
    key=None,
    operator=None,
    geometry=None,
    crop=None,
    # store vs return
    store=None,
    returnas=None,
    return_param=None,
    # specific to deg = 0
    centered=None,
    # to return gradR, gradZ, for D1N2 deg 0, for tomotok
    returnas_element=None,
):

    # ---------
    # check

    (
        key, store, returnas,
        crop, cropbs, cropbs_flat, keycropped,
    ) = _check(
        coll=coll,
        key=key,
        operator=operator,
        geometry=geometry,
        crop=crop,
        store=store,
        returnas=returnas,
    )

    # -------
    # compute

    dout = _get_bsplines_operator(
        coll=coll,
        key=key,
        operator=operator,
        geometry=geometry,
        crop=crop,
        cropbs=cropbs,
        cropbs_flat=cropbs_flat,
        keycropped=keycropped,
        store=store,
        returnas=returnas,
        # specific to deg = 0
        centered=centered,
        # to return gradR, gradZ, for D1N2 deg 0, for tomotok
        returnas_element=returnas_element,
    )

    # ------
    # store

    if store is True:

        for k0, v0 in dout.items():

            if operator == 'D1' and None in v0['ref']:
                continue

            coll.add_data(**v0)

    # return
    if returnas is True:

        if return_param:
            dpar = {
                'key': key,
                'keys': list(dout.keys()),
                'operator': operator,
                'geometry': geometry,
                'crop': crop,
            }
            return dout, dpar

        return dout


# ################################################################
# ################################################################
#               check
# ################################################################


def _check(
    coll=None,
    key=None,
    operator=None,
    geometry=None,
    crop=None,
    store=None,
    returnas=None,
):

    # --------
    # key

    wm = coll._which_mesh
    wbs = coll._which_bsplines
    lk = list(coll.dobj.get(wbs, {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lk,
    )
    keym = coll.dobj[wbs][key][wm]
    nd = coll.dobj[wm][keym]['nd']
    mtype = coll.dobj[wm][keym]['type']

    # --------
    # store

    store = ds._generic_check._check_var(
        store, 'store',
        default=True,
        types=bool,
    )

    # --------
    # returnas

    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        default=store is False,
        types=bool,
    )

    # --------
    # crop

    crop = ds._generic_check._check_var(
        crop, 'crop',
        default=True,
        types=bool,
    )
    if not (nd == '2d' and mtype == 'rect'):
        crop = False

    # cropbs
    cropbs = coll.dobj['bsplines'][key]['crop']
    keycropped = coll.dobj['bsplines'][key]['ref-bs'][0]
    if cropbs not in [None, False] and crop is True:
        cropbs_flat = coll.ddata[cropbs]['data'].ravel(order='F')
        if coll.dobj['bsplines'][key]['deg'] == 0:
            cropbs = coll.ddata[cropbs]['data']
        keycropped = f"{keycropped}-crop"
    else:
        cropbs = False
        cropbs_flat = False

    return key, store, returnas, crop, cropbs, cropbs_flat, keycropped


# ################################################################
# ################################################################
#                       compute
# ################################################################


def _get_bsplines_operator(
    coll=None,
    key=None,
    operator=None,
    geometry=None,
    crop=None,
    cropbs=None,
    cropbs_flat=None,
    keycropped=None,
    store=None,
    returnas=None,
    # specific to deg = 0
    centered=None,
    # to return gradR, gradZ, for D1N2 deg 0, for tomotok
    returnas_element=None,
):

    # -------------------
    # compute and return

    (
        opmat, operator, geometry,
    ) = coll.dobj['bsplines'][key]['class'].get_operator(
        operator=operator,
        geometry=geometry,
        cropbs_flat=cropbs_flat,
        # specific to deg=0
        cropbs=cropbs,
        centered=centered,
        # to return gradR, gradZ, for D1N2 deg 0, for tomotok
        returnas_element=returnas_element,
    )

    # -----------
    # format dout

    wm = coll._which_mesh
    wbs = coll._which_bsplines
    keym = coll.dobj[wbs][key][wm]
    nd = coll.dobj[wm][keym]['nd']
    mtype = coll.dobj[wm][keym]['type']
    deg = coll.dobj[wbs][key]['deg']

    dout = _dout(
        coll=coll,
        key=key,
        opmat=opmat,
        operator=operator,
        geometry=geometry,
        keycropped=keycropped,
        crop=crop,
        nd=nd,
        mtype=mtype,
        deg=deg,
    )

    return dout


# ###################################################
# ###################################################
#                   ref
# ###################################################


def _dout(
    coll=None,
    key=None,
    opmat=None,
    operator=None,
    geometry=None,
    keycropped=None,
    crop=None,
    nd=None,
    mtype=None,
    deg=None,
):

    # --------
    # get refs

    ref, units = _ref_units(
        coll=coll,
        key=key,
        opmat=opmat,
        operator=operator,
        geometry=geometry,
        keycropped=keycropped,
        nd=nd,
        mtype=mtype,
        deg=deg,
    )

    geom = geometry[:3]

    # ----------
    # matrix types

    nnd = int(nd[0])
    if operator == 'D1':
        kmat = [f'M{ii}' for ii in range(nnd)]
    if operator in ['D0N1']:
        kmat = 'M'
    elif operator in ['D0N2']:
        kmat = 'tMM'
    elif operator in ['D1N2']:
        kmat = [f'tMM{ii}' for ii in range(nnd)]
    elif operator in ['D2N2']:
        lcomb = [] if nd == '1d' else [(0,1)]
        kmat = (
            [f'tMM{ii}{ii}' for ii in range(nnd)]
            + [f'tMM{ii}{jj}' for ii, jj in lcomb]
        )

    # ----------
    # build dout

    dout = {}
    if operator in ['D0N1', 'D0N2']:

        k0 = f'{key}_{operator}_{geom}'
        if crop is True:
            k0 = f'{k0}_crop'

        dout[kmat] = {
            'key': k0,
            'data': opmat,
            'ref': ref,
            'units': units[0],
            'dim': operator,
        }

    elif operator in ['D1', 'D1N2']:

        for ii in range(nnd):

            k0 = f'{key}_{operator}_d{ii}'
            if 'N' in operator:
                k0 = f'{k0}_{geom}'
            if crop is True:
                k0 = f'{k0}_crop'

            dout[kmat[ii]] = {
                'key': k0,
                'data': opmat[ii],
                'ref': ref,
                'units': units[ii],
                'dim': operator,
            }

    elif operator in ['D2N2']:

        for ii, kk in enumerate(kmat):

            k0 = f'{key}_{operator}_d{kk[-2:]}_{geom}'
            if crop is True:
                k0 = f'{k0}_crop'

            dout[kk] = {
                'key': k0,
                'data': opmat[ii],
                'ref': ref,
                'units': units[ii],
                'dim': operator,
            }

    return dout


def _ref_units(
    coll=None,
    key=None,
    opmat=None,
    operator=None,
    geometry=None,
    keycropped=None,
    nd=None,
    mtype=None,
    deg=None,
):

    # --------
    # prepare

    wm = coll._which_mesh
    wbs = coll._which_bsplines
    keym = coll.dobj[wbs][key][wm]

    ref = keycropped
    ref0 = coll.dobj[wbs][key]['ref-bs']

    if deg > 0:
        kbsm1 = f'{keym}_bs{deg-1}'
        if kbsm1 in coll.dobj[wbs].keys():
            rm1 = coll.dobj[wbs][kbsm1]['ref-bs']
            if ref0 != ref:
                rm1 = f'{rm1}_crop'
        else:
            rm1 = None
    else:
        rm1 = ref

    # ----------
    # ref

    if operator == 'D1':
        ref = (rm1, ref)

    elif operator == 'D0N1':
        ref = (ref,)

    elif 'N2' in operator:
        ref = (ref, ref)

    # --------
    # units

    apex = coll.dobj[wbs][key]['apex']
    u0 = coll.ddata[apex[0]]['units']
    if nd == '1d':
        units = [_units(u0, operator, geometry)]

    else:
        u1 = coll.ddata[apex[1]]['units']
        units0 = _units(u0, operator, geometry)
        units1 = _units(u1, operator, 'linear')

        if operator in ['D0N1', 'D0N2']:
            units = [units0 * units1]

        elif operator in ['D1', 'D1N2']:
            units = [units0, units1]

        elif operator in ['D2N2']:
            units = [units0, units0*u0*units1*u1, units1]

    return ref, units


def _units(u0=None, operator=None, geometry=None):

    if u0 == '':
        u0 = asunits.Unit('')

    if operator == 'D1':
        units = asunits.Unit(1/u0)

    elif operator == 'D0N1':
        if geometry == 'linear':
            units = u0
        else:
            units = u0**2

    elif operator == 'D0N2':
        if geometry == 'linear':
            units = u0
        else:
            units = u0**2

    elif operator == 'D1N2':
        if geometry == 'linear':
            units = asunits.Unit(1/u0)
        else:
            units = asunits.Unit('')

    elif operator == 'D2N2':
        if geometry == 'linear':
            units = asunits.Unit(1/u0**3)
        else:
            units = asunits.Unit(1/u0**2)

    return units



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import scipy.sparse as scpsp
import datastock as ds


from . import _utils_bsplines_operators as _operators


_LOPERATORS_INT = [
    'D1',
    'D2',
    'D3',
    'D0N1',
    'D0N2',
    'D1N2',
    'D2N2',
    'D3N2',
]


# ###############################################################
# ###############################################################
#                   Mesh2DRect - bsplines - operators
# ###############################################################


def get_operators(
    operator=None,
    geometry=None,
    deg=None,
    knots_mult=None,
    knots_per_bs=None,
    overlap=None,
    sparse_fmt=None,
    # specific to deg = 0
    centered=None,
    # for D1N2 only (to feed to tomotok / Mfr)
    returnas_element=None,
):

    # ------------
    # check inputs

    (
        operator, geometry, sparse_fmt, returnas_element,
    ) = _check(
        deg=deg,
        operator=operator,
        geometry=geometry,
        sparse_fmt=sparse_fmt,
        returnas_element=returnas_element,
    )

    # ------------
    # prepare

    nbs = knots_per_bs.shape[1]
    kn = knots_per_bs
    shape = (nbs, nbs)
    indbs = np.arange(0, nbs)

    if 'N2' in operator and deg >= 1:
        # get intersection indices array
        nbtot = np.sum(overlap >= 0)

        # prepare data and indices arrays
        if operator == 'D0N2':
            data = np.full((nbtot,), np.nan)
            row = np.zeros((nbtot,), dtype=int)
            column = np.zeros((nbtot,), dtype=int)
        elif operator == 'D1N2':
            datadx = np.full((nbtot,), np.nan)
            row = np.zeros((nbtot,), dtype=int)
            column = np.zeros((nbtot,), dtype=int)
        elif operator == 'D2N2':
            datad2x = np.full((nbtot,), np.nan)
            row = np.zeros((nbtot,), dtype=int)
            column = np.zeros((nbtot,), dtype=int)

    # ------------
    # D0 - integral

    geom = geometry
    if operator == 'D0N1':

        if deg == 0:

            opmat = _operators._D0N1_Deg0(kn, geom)

        elif deg == 1:

            opmat = _operators._D0N1_Deg1(kn, geom)

        elif deg == 2:

            in1, in21, in22, in3 = _operators._D0N1_Deg2(kn, geom)
            opmat = (in1 + in21 + in22 + in3)

        elif deg == 3:

            msg = "Integral D0N1 not implemented for deg=3 yet!"
            raise NotImplementedError(msg)

    # ------------
    # D1 - gradient

    elif operator == 'D1':

        # Treat separately discrete case
        if deg == 0:
            grad = _operators._D1_Deg0(
                knots_mult=knots_mult,
                nbs=nbs,
                centered=centered,
            )

        elif deg == 1:
            grad = _operators._D1_Deg1(
                knots_mult=knots_mult,
                nbs=nbs,
            )

        elif deg == 2:
            grad = _operators._D1_Deg2(
                knots_mult=knots_mult,
                nbs=nbs,
            )

        elif deg >= 3:
            raise NotImplementedError()

        opmat = (scpsp.csc_matrix(grad),)

    # ------------
    # D0N2

    elif operator == 'D0N2' and deg == 0:

        ix = _operators._D0N2_Deg0(kn, geom)
        opmat = scpsp.diags(
            [ix],
            [0],
            shape=None,
            format=sparse_fmt,
            dtype=float,
        )

    elif operator == 'D0N2':

        # pre-compute integrals
        if deg == 1:
            ix = _operators._D0N2_Deg1(knots_mult, geom)

        elif deg == 2:
            ix = _operators._D0N2_Deg2(knots_mult, geom)

        elif deg == 3:
            msg = "Integral D0N2 not implemented for deg=3!"
            raise NotImplementedError(msg)

        # set non-diagonal elements
        i0 = 0
        for ii in range(nbs):

            # general case
            overlapi = overlap[:, ii][overlap[:, ii] > ii]

            # diagonal element
            data[i0] = ix[0, ii]
            row[i0] = indbs[ii]
            column[i0] = indbs[ii]
            i0 += 1

            # non-diagonal elements (symmetric)
            for jj in overlapi:

                # store (i, j) and (j, i) (symmetric matrix)
                iix = ix[jj - ii, ii]

                data[i0:i0+2] = iix
                row[i0:i0+2] = (indbs[ii], indbs[jj])
                column[i0:i0+2] = (indbs[jj], indbs[ii])
                i0 += 2

        assert i0 == nbtot
        opmat = scpsp.csc_matrix(
            (data, (row, column)),
            shape=shape,
        )

    # ------------
    # D1N2

    elif operator == 'D1N2':

        # Treat separately discrete case
        if deg == 0:
            grad = _operators._D1_Deg0(
                knots_mult=knots_mult,
                nbs=nbs,
            )

            # surface elements
            if geom == 'linear':
                dx = (knots_mult[1:] - knots_mult[:-1])
            else:
                dx = 0.5*(knots_mult[1:]**2 - knots_mult[:-1]**2)

            # Does not seem to give positive definite matrix, TBC?
            if returnas_element is True:
                opmat = (scpsp.csc_matrix(
                    grad*np.sqrt(dx[:, None])
                ),)

            else:
                opmat = (scpsp.csc_matrix(
                    (grad.T.dot(grad))*(dx[:, None])
                ),)

        else:

            # pre-compute integrals for exact operator deg >= 1
            if deg == 1:
                idx = _operators._D1N2_Deg1(knots_mult, geom)

            elif deg == 2:
                idx = _operators._D1N2_Deg2(knots_mult, geom)

            elif deg == 3:
                msg = "Integral D1N2 not implemented for deg=3!"
                raise NotImplementedError(msg)

            # set non-diagonal elements
            i0 = 0
            for ii in range(nbs):

                # general case
                overlapi = overlap[:, ii][overlap[:, ii] > ii]

                # diagonal element
                datadx[i0] = idx[0, ii]
                row[i0] = indbs[ii]
                column[i0] = indbs[ii]
                i0 += 1

                # non-diagonal elements (symmetric)
                for jj in overlapi:

                    # store (i, j) and (j, i) (symmetric matrix)
                    iidx = idx[jj - ii, ii]

                    datadx[i0:i0+2] = iidx
                    row[i0:i0+2] = (indbs[ii], indbs[jj])
                    column[i0:i0+2] = (indbs[jj], indbs[ii])
                    i0 += 2

            assert i0 == nbtot
            opmat = (scpsp.csc_matrix(
                (datadx, (row, column)),
                shape=shape,
            ),)

    # ------------
    # D2N2

    elif operator == 'D2N2':

        # pre-compute integrals
        if deg in [0, 1]:
            msg = f"degree {deg} too low for operator {operator}"
            raise Exception(msg)
        if deg == 2:
            id2x = _operators._D2N2_Deg2(knots_mult, geom)
            idx = _operators._D1N2_Deg2(knots_mult, geom)
            ix = _operators._D0N2_Deg2(knots_mult, geom)
        elif deg == 3:
            msg = "Integral D2N2 not implemented for deg=3!"
            raise NotImplementedError(msg)

        # set non-diagonal elements
        i0 = 0
        for ii in range(nbs):

                # general case
                overlapi = overlap[:, ii][overlap[:, ii] > ii]

                # diagonal element
                datad2x[i0] = id2x[0, ii]
                row[i0] = indbs[ii]
                column[i0] = indbs[ii]
                i0 += 1

                # non-diagonal elements (symmetric)
                for jj in overlapi:

                    # store (i, j) and (j, i) (symmetric matrix)
                    iid2x = id2x[jj - ii, ii]

                    datad2x[i0:i0+2] = iid2x
                    row[i0:i0+2] = (indbs[ii], indbs[jj])
                    column[i0:i0+2] = (indbs[jj], indbs[ii])
                    i0 += 2

        assert i0 == nbtot
        opmat = (scpsp.csc_matrix(
            (datad2x, (row, column)),
            shape=shape,
        ),)

    # ------------
    # D3N2

    elif operator == 'D3N2' and deg == 3:

        raise NotImplementedError("Integral D3N2 not implemented for deg=3!")

    return opmat, operator, geometry


# ###############################################################
# ###############################################################
#                   checks
# ###############################################################


def _check(
    deg=None,
    operator=None,
    geometry=None,
    sparse_fmt=None,
    returnas_element=None,
):

    # deg
    deg = ds._generic_check._check_var(
        deg, 'deg',
        types=int,
        allowed=[0, 1, 2, 3],
    )

    # operator
    operator = ds._generic_check._check_var(
        operator, 'operator',
        default='D0N1',
        types=str,
        allowed=_LOPERATORS_INT,
    )

    # geometry
    geometry = ds._generic_check._check_var(
        geometry, 'geometry',
        default='toroidal',
        types=str,
        allowed=['toroidal', 'linear'],
    )

    # sparse_fmt
    sparse_fmt = ds._generic_check._check_var(
        sparse_fmt, 'sparse_fmt',
        default='csc',
        types=str,
        allowed=['dia', 'csr', 'csc', 'lil'],
    )

    # returnas_element
    lok = [False]
    if operator == 'D1N2' and deg == 0:
        lok.append(True)
    returnas_element = ds._generic_check._check_var(
        returnas_element, 'returnas_element',
        default=False,
        types=bool,
        allowed=lok,
    )

    return operator, geometry, sparse_fmt, returnas_element



"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import itertools as itt


# Standard
import numpy as np
import matplotlib.pyplot as plt


# specific
_HERE = os.path.abspath(os.path.dirname(__file__))
_PATH_DATA = os.path.join(_HERE, 'test_data')
_DFNAME = {
    'WEST_tri': 'WEST_trimesh.npz',
    'quad_tri': 'trimesh_quad.npz',
    'WEST_poly': 'WEST_Poly.npz',
}


_DDATA = {}
for k0, fname in _DFNAME.items():
    _PFE = os.path.join(_PATH_DATA, fname)
    dd = dict(np.load(_PFE, allow_pickle=True))
    if k0 == 'WEST':
        _DDATA[k0] = dd
    else:
        _DDATA[k0] = {}
        for k1, v1 in dd.items():
            _DDATA[k0][k1] = v1.tolist()


#######################################################
#
#     Add fixed meshes
#
#######################################################


def _add_1d_knots_uniform(bsplines, key=None):
    bsplines.add_mesh_1d(key=key, knots=np.linspace(0, 10, 11), units='eV')


def _add_1d_knots_variable(bsplines, key=None):
    bsplines.add_mesh_1d(
        key=key, knots=np.r_[0, 1, 4, 7, 10], uniform=False, units='A',
    )


def _add_rect_uniform(bsplines, key=None):
    # add uniform rect mesh
    bsplines.add_mesh_2d_rect(
        key=key, domain=[[2, 3], [-1, 1]], res=0.1, units='GHz',
    )


def _add_rect_variable(bsplines, key=None):
    # add variable rect mesh
    bsplines.add_mesh_2d_rect(
        key=key,
        domain=[[2, 2.3, 2.6, 3], [-1, 0., 1]],
        res=[[0.2, 0.1, 0.1, 0.2], [0.2, 0.1, 0.2]],
        units='m',
    )


def _add_rect_variable_crop(bsplines, key=None):
    # add variable rect mesh
    bsplines.add_mesh_2d_rect(
        key=key,
        domain=[[2, 2.3, 2.6, 3], [-1, 0., 1]],
        res=[[0.2, 0.1, 0.1, 0.2], [0.2, 0.1, 0.2]],
        crop_poly=_DDATA['WEST_poly']['Poly'],
        units='cm',
    )


def _add_rect_crop_from_knots(bsplines, key=None):
    # add variable rect mesh
    bsplines.add_mesh_2d_rect(
        key=key,
        knots0=np.linspace(2, 3, 11),
        knots1=np.linspace(-1, 1, 11),
        crop_poly=_DDATA['WEST_poly']['Poly'],
        units='mm',
    )


def _add_rect_variable_crop_from_knots(bsplines, key=None):
    # add variable rect mesh
    bsplines.add_mesh_2d_rect(
        key=key,
        knots0=np.r_[np.linspace(2, 2.4, 5), np.r_[2.5, 2.7, 2.8, 3.]],
        knots1=np.linspace(-1, 1, 11),
        crop_poly=_DDATA['WEST_poly']['Poly'],
        units='km',
    )


def _add_tri_ntri1(bsplines, key=None):

    knots = np.array([
        _DDATA['WEST_tri']['pts_x0'],
        _DDATA['WEST_tri']['pts_x1'],
    ]).T

    bsplines.add_mesh_2d_tri(
        key=key,
        knots=knots,
        indices=_DDATA['WEST_tri']['indices'],
        units='dm',
    )


def _add_tri_ntri2(bsplines, key=None):

    knots = np.array([
        _DDATA['quad_tri']['pts_x0'],
        _DDATA['quad_tri']['pts_x1'],
    ]).T

    bsplines.add_mesh_2d_tri(
        key=key,
        knots=knots,
        indices=_DDATA['quad_tri']['indices'],
        units='um',
    )


def _add_tri_delaunay(bsplines, key=None):

    bsplines.add_mesh_2d_tri(
        key=key,
        pts_x0=_DDATA['quad_tri']['pts_x0'],
        pts_x1=_DDATA['quad_tri']['pts_x1'],
        units='nm',
    )


#######################################################
#
#     Add variable meshes
#
#######################################################


def _add_mesh_1d_subkey(bs, nd=None, kind=None):

    wbs = bs._which_bsplines
    wm = bs._which_mesh

    dkd = _get_data(bs, nd=nd, kind=kind)
    dkd = {
        k0: v0 for k0, v0 in dkd.items()
        if len(bs.ddata[k0][wbs]) == 1
    }

    ni = 0
    for ii, (k0, v0) in enumerate(dkd.items()):

        kbs = bs.ddata[k0]['bsplines'][0]
        km = bs.dobj[wbs][kbs]['mesh']
        deg = bs.dobj[wbs][kbs]['deg']
        if deg != 2 or ni > 0:
            continue

        nm = len(bs.dobj[wm])
        key = f'm{nm:02.0f}'
        bs.add_mesh_1d(
            key=key,
            knots=np.linspace(0, 20, 5),
            subkey=k0,
        )
        ni += 1

        assert bs.dobj[wm][key]['subkey'] == (k0,)
        assert bs.dobj[wm][key]['submesh'] == km


def _add_mesh_2d_rect_subkey(bs, nd=None, kind=None):

    wbs = bs._which_bsplines
    wm = bs._which_mesh

    dkd = _get_data(bs, nd=nd, kind=kind)
    dkd = {
        k0: v0 for k0, v0 in dkd.items()
        if len(bs.ddata[k0][wbs]) == 1
    }

    ni = 0
    for ii, (k0, v0) in enumerate(dkd.items()):

        kbs = bs.ddata[k0]['bsplines'][0]
        km = bs.dobj[wbs][kbs]['mesh']
        deg = bs.dobj[wbs][kbs]['deg']
        if deg != 2 or ni > 0:
            continue

        nm = len(bs.dobj[wm])
        key = f'm{nm:02.0f}'
        bs.add_mesh_2d_rect(
            key=key,
            knots0=np.linspace(0, 20, 5),
            knots1=np.linspace(0, 20, 5),
            subkey0=k0,
            subkey1=k0,
        )
        ni += 1

        assert bs.dobj[wm][key]['subkey'] == (k0, k0)
        assert bs.dobj[wm][key]['submesh'] == km


def _add_polar1(bsplines, key=None):
    """ Time-independent """

    kR, kZ = bsplines.dobj['bsplines']['m2-bs1']['apex']
    R = bsplines.ddata[kR]['data']
    Z = bsplines.ddata[kZ]['data']
    RR = np.repeat(R[:, None], Z.size, axis=1)
    ZZ = np.repeat(Z[None, :], R.size, axis=0)
    rho = (RR - 2.5)**2/0.08 + (ZZ - 0)**2/0.35

    bsplines.add_data(
        key='rho1',
        data=rho,
        ref='m2-bs1',
        unit='',
        dim='',
        quant='rho',
        name='rho',
    )

    bsplines.add_mesh_2d_polar(
        key=key,
        radius=np.linspace(0, 1.2, 7),
        angle=None,
        radius2d='rho1',
    )


def _add_polar2(bsplines, key=None):
    """ Time-dependent """

    kR, kZ = bsplines.dobj['bsplines']['m2-bs1']['apex']
    R = bsplines.ddata[kR]['data']
    Z = bsplines.ddata[kZ]['data']
    RR = np.repeat(R[:, None], Z.size, axis=1)
    ZZ = np.repeat(Z[None, :], R.size, axis=0)

    rho = (RR - 2.5)**2/0.08 + (ZZ - 0)**2/0.35
    angle = np.arctan2(ZZ/2., (RR - 2.5))

    nt = 11
    t = np.linspace(30, 40, nt)
    rho = rho[None, ...] + 0.1*np.cos(t)[:, None, None]**2
    angle = angle[None, ...] + 0.01*np.sin(t)[:, None, None]**2


    if 'nt' not in bsplines.dref.keys():
        bsplines.add_ref(
            key='nt',
            size=nt,
        )

    if 't' not in bsplines.ddata.keys():
        bsplines.add_data(
            key='t',
            data=t,
            ref=('nt',),
            dim='time',
        )

    if 'rho2' not in bsplines.ddata.keys():
        bsplines.add_data(
            key='rho2',
            data=rho,
            ref=('nt', 'm2-bs1'),
            unit='',
            dim='',
            quant='rho',
            name='rho',
        )

    if 'angle2' not in bsplines.ddata.keys():
        bsplines.add_data(
            key='angle2',
            data=angle,
            ref=('nt', 'm2-bs1'),
            unit='rad',
            dim='',
            quant='angle',
            name='theta',
        )

    # ang
    if key == 'm6':
        ang = np.pi*np.r_[-3./4., -1/4, 0, 1/4, 3/4]
    else:
        ang = None

    # mesh
    bsplines.add_mesh_2d_polar(
        key=key,
        radius=np.linspace(0, 1.2, 7),
        angle=ang,
        radius2d='rho2',
        angle2d='angle2',
    )


#######################################################
#
#     Add bsplines
#
#######################################################


def _add_bsplines(
    bs,
    key=None,
    nd=None,
    kind=None,
    subkey=None,
    deg=None,
    angle=None,
):

    lkm = _get_mesh(bs, nd=nd, kind=kind)
    if isinstance(deg, int):
        deg = [deg]
    elif deg is None:
        deg = [0, 1, 2, 3]

    ddeg = {
        None: [ii for ii in [0, 1, 2, 3] if ii in deg],
        'rect': [ii for ii in [0, 1, 2, 3] if ii in deg],
        'tri': [ii for ii in [0, 1] if ii in deg],
        'polar': [ii for ii in [0, 1, 2, 3] if ii in deg],
    }

    wm = bs._which_mesh
    for km in lkm:
        mtype = bs.dobj[wm][km]['type']
        if subkey is None:
            for dd in ddeg[mtype]:
                bs.add_bsplines(key=km, deg=dd)
        elif bs.dobj[wm][km]['subkey'] is not None:
            bs.add_bsplines(key=km, deg=1)


#######################################################
#
#     Manipulate mesh and bsplines
#
#######################################################


def _select_mesh_elements(bsplines, nd=None, kind=None):
    lkm = _get_mesh(bsplines, nd=nd, kind=kind)

    if nd == '1d':
        lcrop = [None]
        lind = [None, 0, [0, 3]]
        lneigh = [False]

    elif kind == 'rect':
        lcrop = [False, True]
        lind = [None, 0, [0, 3]]
        lneigh = [False, True]

    else:
        lcrop = [None]
        lind = [None, 0, [0, 3]]
        lneigh = [False, True]

    lel = ['knots', 'cents']
    lret = ['ind', 'data']

    for km in lkm:

        for comb in itt.product(lcrop, lel, lind, lret, lneigh):
            out = bsplines.select_mesh_elements(
                key=km,
                ind=comb[2],
                elements=comb[1],
                returnas=comb[3],
                return_neighbours=comb[4],
                crop=comb[0],
            )


def _sample_mesh(bsplines, nd=None, kind=None):
    lkm = _get_mesh(bsplines, nd=nd, kind=kind)

    lres = [0.1, 0.3]
    lmode = ['abs', 'rel']
    lin_mesh = [False, True]
    limshow = [False, True]

    Dx0, Dx1 = None, None
    for km in lkm:
        for ii, comb in enumerate(itt.product(lres, lmode, lin_mesh, limshow)):

            if nd == '2d':
                if comb[1] == 'abs' and ii % 3 == 0:
                    Dx0 = np.r_[2., 3., 2.5]
                    Dx1 = np.r_[-1, -1, 1]
                elif comb[1] == 'abs' and ii % 3 == 1:
                    Dx0 = [None, 2.5]
                    Dx1 = [-0.5, None]
                else:
                    Dx0 = None
                    Dx1 = None

            out = bsplines.get_sample_mesh(
                key=km,
                res=comb[0],
                mode=comb[1],
                grid=None,
                Dx0=Dx0,
                Dx1=Dx1,
                in_mesh=comb[2] and kind != 'tri',
                imshow=comb[3]
            )


def _plot_mesh(bsplines, nd=None, kind=None):
    lkm = _get_mesh(bsplines, nd=nd, kind=kind)

    lk = [None, 2, [2, 3]]
    lc = [None, 2, [2, 3]]
    if kind == 'rect':
        lcrop = [False, True]
    else:
        lcrop = [False]

    for km in lkm:

        for comb in itt.product(lk, lc, lcrop):
            _ = bsplines.plot_mesh(
                key=km,
                ind_knot=comb[0],
                ind_cent=comb[1],
                crop=comb[2],
            )
            plt.close('all')


#######################################################
#
#     Add data on bsplines
#
#######################################################


def _select_bsplines(bs, nd=None, kind=None):
    lkb = _get_bs(bs, nd=nd, kind=kind)

    lind = [None, 0, [0, 3]]
    if kind == 'rect':
        lcrop = [False, True]
    else:
        lcrop = [None]
    lretc = [True, False]
    lretk = [True, False]

    for km in lkb:

        for comb in itt.product(lind, lretc, lretk, lcrop):
            out = bs.select_bsplines(
                key=km,
                ind=comb[0],
                return_cents=comb[1],
                return_knots=comb[2],
                crop=comb[3],
            )

def _add_data_1bs_fix(bs, nd=None, kind=None, remove=None):
    lkb = _get_bs(bs, nd=nd, kind=kind)

    lkd = []
    for kb in lkb:

        lkd.append(f'{kb}_fix')
        shape = bs.dobj[bs._which_bsplines][kb]['shape']
        data = np.random.random(shape)

        bs.add_data(
            key=lkd[-1],
            data=data,
            ref=kb,
        )

        assert bs.ddata[lkd[-1]][bs._which_bsplines] == (kb,)
        bsref = bs.dobj[bs._which_bsplines][kb]['ref']
        assert bs.ddata[lkd[-1]]['ref'] == bsref

    if remove:
        for kd in lkd:
            bs.remove_data(kd)


def _add_data_1bs_arrays(bs, nd=None, kind=None, remove=None):
    lkb = _get_bs(bs, nd=nd, kind=kind)

    nt, nE = 10, 11
    if 'nt' not in bs.dref.keys():
        bs.add_ref(key='nt', size=nt)
        bs.add_ref(key='nE', size=nE)
        bs.add_data(key='t', data=np.linspace(1, 1.1, nt), ref='nt', unit='s')

    lkd = []
    for kb in lkb:

        lkd.append(f'{kb}_fix')
        shape = np.r_[nt, bs.dobj[bs._which_bsplines][kb]['shape'], nE]
        data = np.random.random(shape)

        bs.add_data(
            key=lkd[-1],
            data=data,
            ref=['nt', kb, 'nE'],
        )

        assert bs.ddata[lkd[-1]][bs._which_bsplines] == (kb,)
        bsref = bs.dobj[bs._which_bsplines][kb]['ref']
        ref = tuple(['nt'] + list(bsref) + ['nE'])
        assert bs.ddata[lkd[-1]]['ref'] == ref

    if remove:
        for kd in lkd:
            bs.remove_data(kd)


def _add_data_multibs_arrays(bs, nd=None, kind=None, subbs=None, remove=None):
    lkb = _get_bs(bs, nd=nd, kind=kind, subbs=subbs)

    nt, nE = 10, 11
    if 'nt' not in bs.dref.keys():
        bs.add_ref(key='nt', size=nt)
        bs.add_ref(key='nE', size=nE)
        bs.add_data(key='t', data=np.linspace(1, 1.1, nt), ref='nt', unit='s')

    lkd = []
    wbs = bs._which_bsplines
    for ii, kb in enumerate(lkb):

        lkd.append(f'{kb}_var')

        kb2 = lkb[(ii + int(len(lkb)/2)) % len(lkb)]
        shape = np.r_[
            nt,
            bs.dobj[wbs][kb]['shape'],
            nE,
            bs.dobj[wbs][kb2]['shape'],
        ]
        data = np.random.random(shape)

        bs.add_data(
            key=lkd[-1],
            data=data,
            ref=['nt', kb, 'nE', kb2],
            units='ph/(m3.sr.eV.s)',
        )

        if bs.ddata[lkd[-1]][wbs] != (kb, kb2):
            msg = (
                f"Wrong '{wbs}' for ddata['{lkd[-1]}']:\n"
                f"{bs.ddata[lkd[-1]][wbs]} vs {(kb, kb2)}"
            )
            raise Exception(msg)

        bsref = bs.dobj[wbs][kb]['ref']
        bsref2 = bs.dobj[wbs][kb2]['ref']
        ref = tuple(['nt'] + list(bsref) + ['nE'] + list(bsref2))
        if bs.ddata[lkd[-1]]['ref'] != ref:
            msg = (
                f"Wrong '{wbs}' for ddata['{lkd[-1]}']:\n"
                f"{bs.ddata[lkd[-1]]['ref']} vs {ref}"
            )
            raise Exception(msg)

    if remove:
        for kd in lkd:
            bs.remove_data(kd)


#######################################################
#
#     Binning and interpolation
#
#######################################################


def _interpolate(bs, nd=None, kind=None, details=None, submesh=None):
    dkd = _get_data(
        bs, nd=nd, kind=kind, submesh=submesh, maxref=int(nd[0]) + 2,
    )

    wm = bs._which_mesh
    wbs = bs._which_bsplines
    for ii, (kd, vd) in enumerate(dkd.items()):

        # crop
        ref_key = vd['bs'][0]
        if kind == 'rect' and isinstance(bs.dobj[wbs][ref_key]['crop'], str):
            crop = (ii % 2 == 0)
        else:
            crop = False

        # vect
        if submesh is True:
            km = bs.dobj[wbs][ref_key][wm]
            kd0 = bs.dobj[wm][km]['subkey']
            kbs0 = bs.dobj[wm][km]['subbs']
            km0 = bs.dobj[wm][km]['submesh']
            nd0 = bs.dobj[wm][km0]['nd']
            vect = bs.dobj[wbs][kbs0]['apex'][0]
        else:
            vect = bs.dobj[wbs][ref_key]['apex'][0]
            nd0 = None

        vect = np.r_[bs.ddata[vect]['data'][0], bs.ddata[vect]['data'][-1]]
        vect = np.array([vect, vect])

        # add x0, x1
        if 'nt' in bs.ddata[kd]['ref'] and (ii % 4 != 0):

            vect = np.tile(vect, (bs.dref['nt']['size'], 1, 1))

            if 'n2' not in bs.dref.keys():
                bs.add_ref(key='n2', size=2)

            bs.add_data(
                key='x0',
                data=vect,
                ref=('nt', 'n2', 'n2'),
                units='blabla',
            )
            bs.add_data(
                key='x1',
                data=vect,
                ref=('nt', 'n2', 'n2'),
                units='blabla',
            )
            x0 = 'x0'
            x1 = 'x1'

            # ref_com
            if kind != 'tri' and (ii % 3 != 0):
                if submesh is True:
                    if 'nt' in bs.ddata[kd0[0]]['ref']:
                        ref_com = 'nt'
                    else:
                        ref_com = None
                else:
                    ref_com = 'nt'
            else:
                ref_com = None

        else:
            x0 = vect
            x1 = vect
            ref_com = None

        # populate
        kwd = {'x0': x0, 'ref_com': ref_com}
        if nd == '2d' or nd0 == '2d':
            kwd['x1'] = x1

        # indbs_tf
        shapebs = bs.dobj[bs._which_bsplines][ref_key]['shape']
        if kind == 'rect':
            if ii%3 == 0:
                indbs_tf = bs.select_ind(
                    key=ref_key,
                    ind=([1, 2, 5], [4, 5, 3]),
                    crop=crop,
                    returnas='tuple-flat',
                )
                nbs = len(indbs_tf[0])
            else:
                indbs_tf = None
                if crop is True:
                    kcrop = bs.dobj[wbs][ref_key]['crop']
                    nbs = np.sum(bs.ddata[kcrop]['data'])
                else:
                    nbs = np.prod(shapebs)
        else:
            if ii%3 == 0:
                indbs_tf = np.arange(0, np.prod(shapebs), 2)
                nbs = indbs_tf.size
            else:
                indbs_tf = None
                nbs = np.prod(shapebs)

        kwd.update({
            'domain': None,
            'details': details,
            'indbs_tf': indbs_tf,
            'crop': crop,
            'submesh': submesh,
        })

        # interpolate
        dout, dparam = bs.interpolate(
            keys=None if details is True else (None if ii%2 == 0 else kd),
            ref_key=ref_key,
            return_params=True,
            debug=True,
            **kwd,
        )

        # expected shape
        if details is True:
            shape = tuple(np.r_[vect.shape, nbs].astype(int))
            kd = f'{ref_key}_details'
        else:
            shape = list(bs.ddata[kd]['shape'])
            ax0 = dparam['daxis'][kd][0]
            ax1 = dparam['daxis'][kd][-1]

            # submesh
            if submesh is True:
                refbs0 = bs.dobj[wbs][kbs0]['ref']
                ax00 = bs.ddata[kd0[0]]['ref'].index(refbs0[0])
                ax10 = bs.ddata[kd0[0]]['ref'].index(refbs0[-1])
                shape0 = bs.ddata[kd0[0]]['shape']

            if ref_com is None:
                if submesh is True:
                    vshape = tuple(np.r_[
                        shape0[:ax00], vect.shape, shape0[ax10+1:]
                    ].astype(int))
                    shape = tuple(np.r_[
                        shape[:ax0], vshape, shape[ax1+1:]
                    ].astype(int))

                else:
                    shape = tuple(np.r_[
                        shape[:ax0], vect.shape, shape[ax1+1:]
                    ].astype(int))
            else:
                if submesh is True:
                    vshape = tuple(np.r_[
                        shape0[:ax00], vect.shape[1:], shape0[ax10+1:]
                    ].astype(int))
                    shape = tuple(np.r_[
                        shape[:ax0], vshape, shape[ax1+1:]
                    ].astype(int))
                else:
                    shape = tuple(np.r_[
                        shape[:ax0], vect.shape[1:], shape[ax1+1:]
                    ].astype(int))

        # remove x0, x1
        if 'x0' in bs.ddata.keys():
            bs.remove_data('x0')
            bs.remove_data('x1')

        # error msg
        if dout[kd]['data'].shape != shape:
            lstr = [
                f"\t- {k0}: {v0}"
                for k0, v0 in dparam.items()
                if k0 not in ['x0', 'x1']
            ]
            msg = (
                f"Wrong shape for '{kd}':\n"
                f"\t- expected: {shape}\n"
                f"\t- obtained: {dout[kd]['data'].shape}\n\n"
                f"\t- nd: {nd}\n"
                f"\t- kind: {kind}\n"
                f"\t- ddata['{kd}']['data'].shape: {bs.ddata[kd]['shape']}\n"
                f"\t- ddata['{kd}']['ref'].shape: {bs.ddata[kd]['ref']}\n\n"
                + "\n".join(lstr)
            )
            raise Exception(msg)


def _bin_bs(bs, nd=None, kind=None):
    dkd = _get_data(bs, nd=nd, kind=kind, maxref=3)
    wbs = bs._which_bsplines
    for ii, (kd, vd) in enumerate(dkd.items()):


        if len(bs.ddata[kd]['ref']) > 3:
            continue

        if len(vd['ref']) > 1:
            if ii % 10 == 0:
                ref_key = vd['ref'][0]
                ax = 0
            else:
                ref_key = vd['bs'][0]
                ax = bs.ddata[kd]['ref'].index(bs.dobj[wbs][ref_key]['ref'][0])
        else:
            ref_key = vd['bs'][0]
            ax = bs.ddata[kd]['ref'].index(bs.dobj[wbs][ref_key]['ref'][0])

        vect = bs.dobj[bs._which_bsplines][vd['bs'][0]]['apex'][0]
        vect = bs.ddata[vect]['data']
        if len(vect) > 6:
            vect = vect[::3]

        dd = np.abs(np.mean(np.diff(vect)))
        DD = vect[-1] - vect[0]
        nbins = int(DD/dd)
        bins = np.linspace(vect[0] - 0.1*DD, vect[0]+0.5*DD, nbins)

        dout = bs.binning(
            data=kd,
            bin_data0=ref_key,
            bins0=bins,
            # store vs return
            store=False,
            returnas=True,
        )
        kd = list(dout.keys())[0]

        shape = list(bs.ddata[kd]['shape'])
        shape[ax] = nbins
        shape = tuple(shape)
        if dout[kd]['data'].shape != shape:
            shd = bs.ddata[kd]['data'].shape
            msg = (
                "Binning of data '{kd}' along ref 'ref_key' has wrong shape:\n"
                f"\t- ddata['{kd}']['data'].shape = {shd}\n"
                f"\t- dout['{kd}']['data'].shape = {dout[kd]['data'].shape}\n"
                f"\t- expected shape = {shape}\n"
            )
            raise Exception(msg)


def _add_data_var(bsplines, key):

    kdata = f'{key}-data-var'
    shape = bsplines.dobj['bsplines'][key]['shape']
    t = bsplines.ddata['t']['data']
    tsh = tuple([t.size] + [1 for ii in shape])
    data = np.cos(t.reshape(tsh)) * np.random.random(shape)[None, ...]

    bsplines.add_data(
        key=kdata,
        data=data,
        ref=('nt', key),
    )
    return kdata


#######################################################
#
#     Plotting
#
#######################################################


def _plot_as_profile2d(bs,  nd=None, kind=None):
    dkd = _get_data(bs, nd=nd, kind=kind, maxref=4)

    wm = bs._which_mesh
    wbs = bs._which_bsplines

    for ii, (k0, v0) in enumerate(dkd.items()):

        if bs.ddata[k0]['data'].ndim > 4:
            continue
        if k0 in bs.dobj[wbs][v0['bs'][0]]['apex']:
            continue
        if ii%2 == 0:
            continue

        # knots
        knots = bs.dobj[wbs][v0['bs'][0]]['apex'][0]
        knots = bs.ddata[knots]['data']
        span = np.abs(knots[-1] - knots[0])
        res = span / 3.

        dax = bs.plot_as_profile2d(
            key=k0,
            dres=res,
        )
        plt.close('all')


#######################################################
#
#     Operators
#
#######################################################


def _get_operators(bs, nd=None, kind=None, remove=None):
    lkb = _get_bs(bs, nd=nd, kind=kind)

    operator = [
        'D1',
        'D0N1',
        'D0N2', 'D1N2', 'D2N2', 'D3N2',
    ]
    geometry = ['linear', 'toroidal']
    wm = bs._which_mesh
    wbs = bs._which_bsplines

    for ii, k0 in enumerate(lkb):

        keym = bs.dobj[wbs][k0][wm]
        mtype = bs.dobj[wm][keym]['type']

        for op in operator:

            deg = bs.dobj[wbs][k0]['deg']
            if deg == 0 and op == 'D1N2':
                pass
            elif deg < int(op[1]):
                continue
            elif len(op) == 2 and deg >= 2:
                continue
            elif op == 'D1' and mtype == 'rect' and deg >= 1:
                continue
            elif deg == 3:
                continue

            for gg in geometry:
                if op == 'D1' and gg == 'toroidal':
                    continue
                # print('\t', k0, nd, kind, op, gg)

                bs.add_bsplines_operator(
                    key=k0,
                    operator=op,
                    geometry=gg,
                    store=ii % 2 == 0,
                )


#######################################################
#
#     utilities
#
#######################################################


def _get_mesh(bsplines, nd=None, kind=None):
    return [
        k0 for k0, v0 in bsplines.dobj[bsplines._which_mesh].items()
        if nd in [None, v0['nd']]
        and kind in [None, v0['type']]
    ]


def _get_bs(bs, nd=None, kind=None, subbs=None):
    wm = bs._which_mesh
    wbs = bs._which_bsplines
    return [
        k0 for k0, v0 in bs.dobj[wbs].items()
        if nd in [None, bs.dobj[wm][v0[wm]]['nd']]
        and kind in [None, bs.dobj[wm][v0[wm]]['type']]
        and (
            subbs in [None, bs.dobj[wm][v0[wm]]['subbs']]
            or (subbs is True and bs.dobj[wm][v0[wm]]['subbs'] is not None)
        )
    ]


def _get_data(bs, nd=None, kind=None, submesh=None, maxref=None):

    dkd = {}
    wm = bs._which_mesh
    wbs = bs._which_bsplines
    dbs, dref = bs.get_dict_bsplines()
    for k0 in sorted(dbs.keys()):
        v0 = dbs[k0]
        for kb in v0.keys():
            km = bs.dobj[wbs][kb][wm]
            lc = [
                nd in [None, bs.dobj[wm][km]['nd']],
                kind in [None, bs.dobj[wm][km]['type']],
                (
                    submesh is None
                    or (
                        submesh is True
                        and bs.dobj[wm][km]['submesh'] is not None
                    )
                ),
                maxref is None or len(bs.ddata[k0]['ref']) <= maxref,
            ]
            if all(lc):
                if k0 not in dkd.keys():
                    dkd[k0] = {'bs': [kb], 'ref': dref[k0]}
                else:
                    dkd[k0]['bs'].append(kb)

    return dkd
